from flask import Blueprint, jsonify, request

from fedn.network.api.auth import jwt_auth_required
from fedn.network.api.v1.shared import api_version, get_post_data_to_kwargs, get_typed_list_headers, mdb
from fedn.network.storage.statestore.stores.round_store import RoundStore
from fedn.network.storage.statestore.stores.shared import EntityNotFound

bp = Blueprint("round", __name__, url_prefix=f"/api/{api_version}/rounds")

round_store = RoundStore(mdb, "control.rounds")


@bp.route("/", methods=["GET"])
@jwt_auth_required(role="admin")
def get_rounds():
    """Get rounds
    Retrieves a list of rounds based on the provided parameters.
    By specifying a parameter in the url, you can filter the rounds based on that parameter,
    and the response will contain only the rounds that match the filter.
    ---
    tags:
        - Rounds
    parameters:
      - name: status
        in: query
        required: false
        type: string
        description: Status of the round
      - name: round_id
        in: query
        required: false
        type: string
      - name: X-Limit
        in: header
        required: false
        type: integer
        description: The maximum number of rounds to retrieve
      - name: X-Skip
        in: header
        required: false
        type: integer
        description: The number of rounds to skip
      - name: X-Sort-Key
        in: header
        required: false
        type: string
        description: The key to sort the rounds by
      - name: X-Sort-Order
        in: header
        required: false
        type: string
        description: The order to sort the rounds in ('asc' or 'desc')
    definitions:
      Round:
        type: object
        properties:
          id:
            type: string
          status:
            type: string
          round_id:
            type: string
          round_config:
            type: object
          round_data:
            type: object
          combiners:
            type: array
            description: List of combiner objects used for the round
            items:
              type: object
    responses:
      200:
        description: A list of rounds and the total count.
        schema:
            type: object
            properties:
                count:
                    type: integer
                result:
                    type: array
                    items:
                        $ref: '#/definitions/Round'
      500:
        description: An error occurred
        schema:
            type: object
            properties:
                message:
                    type: string
    """
    try:
        limit, skip, sort_key, sort_order, _ = get_typed_list_headers(request.headers)

        kwargs = request.args.to_dict()

        rounds = round_store.list(limit, skip, sort_key, sort_order, use_typing=False, **kwargs)

        result = rounds["result"]

        response = {"count": rounds["count"], "result": result}

        return jsonify(response), 200
    except Exception:
        return jsonify({"message": "An unexpected error occurred"}), 500


@bp.route("/list", methods=["POST"])
@jwt_auth_required(role="admin")
def list_rounds():
    """List rounds
    Retrieves a list of rounds based on the provided parameters.
    Works much like the GET /rounds endpoint, but allows for more complex queries.
    By specifying a parameter in the request body, you can filter the rounds based on that parameter,
    and the response will contain only the rounds that match the filter. If the parameter value contains a comma,
    the filter will be an "in" query, meaning that the rounds will be returned if the specified field contains any of the values in the parameter.
    ---
    tags:
      - Rounds
    parameters:
      - name: round
        in: body
        required: false
        schema:
          type: object
          properties:
            status:
              type: string
            round_id:
              type: string
      - name: X-Limit
        in: header
        required: false
        type: integer
        description: The maximum number of rounds to retrieve
      - name: X-Skip
        in: header
        required: false
        type: integer
        description: The number of rounds to skip
      - name: X-Sort-Key
        in: header
        required: false
        type: string
        description: The key to sort the rounds by
      - name: X-Sort-Order
        in: header
        required: false
        type: string
        description: The order to sort the rounds in ('asc' or 'desc')
    responses:
      200:
        description: A list of rounds and the total count.
        schema:
            type: object
            properties:
                count:
                    type: integer
                result:
                    type: array
                    items:
                        $ref: '#/definitions/Round'
      500:
        description: An error occurred
        schema:
            type: object
            properties:
                message:
                    type: string
    """
    try:
        limit, skip, sort_key, sort_order, _ = get_typed_list_headers(request.headers)

        kwargs = get_post_data_to_kwargs(request)

        rounds = round_store.list(limit, skip, sort_key, sort_order, use_typing=False, **kwargs)

        result = rounds["result"]

        response = {"count": rounds["count"], "result": result}

        return jsonify(response), 200
    except Exception:
        return jsonify({"message": "An unexpected error occurred"}), 500


@bp.route("/count", methods=["GET"])
@jwt_auth_required(role="admin")
def get_rounds_count():
    """Rounds count
    Retrieves the count of rounds based on the provided parameters.
    By specifying a parameter in the url, you can filter the rounds based on that parameter,
    and the response will contain only the count of rounds that match the filter.
    ---
    tags:
        - Rounds
    parameters:
      - name: round_id
        in: query
        required: false
        type: string
      - name: status
        in: query
        required: false
        type: string
    responses:
        200:
            description: The count of rounds
            schema:
                type: integer
        500:
            description: An error occurred
            schema:
                type: object
                properties:
                    message:
                        type: string
    """
    try:
        kwargs = request.args.to_dict()
        count = round_store.count(**kwargs)
        response = count
        return jsonify(response), 200
    except Exception:
        return jsonify({"message": "An unexpected error occurred"}), 500


@bp.route("/count", methods=["POST"])
@jwt_auth_required(role="admin")
def rounds_count():
    """Rounds count
    Retrieves the count of rounds based on the provided parameters.
    Works much like the GET /rounds/count endpoint, but allows for more complex queries.
    By specifying a parameter in the request body, you can filter the rounds based on that parameter,
    if the parameter value contains a comma, the filter will be an "in" query, meaning that the rounds
    will be returned if the specified field contains any of the values in the parameter.
    ---
    tags:
        - Rounds
    parameters:
      - name: round
        in: body
        required: false
        schema:
          type: object
          properties:
            status:
              type: string
            round_id:
              type: string
    responses:
        200:
            description: The count of rounds
            schema:
                type: integer
        500:
            description: An error occurred
            schema:
                type: object
                properties:
                    message:
                        type: string
    """
    try:
        kwargs = get_post_data_to_kwargs(request)
        count = round_store.count(**kwargs)
        response = count
        return jsonify(response), 200
    except Exception:
        return jsonify({"message": "An unexpected error occurred"}), 500


@bp.route("/<string:id>", methods=["GET"])
@jwt_auth_required(role="admin")
def get_round(id: str):
    """Get round
    Retrieves a round based on the provided id.
    ---
    tags:
        - Rounds
    parameters:
      - name: id
        in: path
        required: true
        type: string
        description: The id of the round
    responses:
        200:
            description: The round
            schema:
                $ref: '#/definitions/Round'
        404:
            description: The round was not found
            schema:
                type: object
                properties:
                    error:
                        type: string
        500:
            description: An error occurred
            schema:
                type: object
                properties:
                    message:
                        type: string
    """
    try:
        round = round_store.get(id, use_typing=False)
        response = round

        return jsonify(response), 200
    except EntityNotFound:
        return jsonify({"message": f"Entity with id: {id} not found"}), 404
    except Exception:
        return jsonify({"message": "An unexpected error occurred"}), 500



from .client_cmd import client_cmd  # noqa: F401
from .combiner_cmd import combiner_cmd  # noqa: F401
from .config_cmd import config_cmd  # noqa: F401
from .main import main  # noqa: F401
from .model_cmd import model_cmd  # noqa: F401
from .package_cmd import package_cmd  # noqa: F401
from .round_cmd import round_cmd  # noqa: F401
from .run_cmd import run_cmd  # noqa: F401
from .session_cmd import session_cmd  # noqa: F401
from .status_cmd import status_cmd  # noqa: F401
from .validation_cmd import validation_cmd  # noqa: F401
from .controller_cmd import controller_cmd  # noqa: F401



import base64
import copy
import json
import time
from io import BytesIO

import grpc

import fedn.network.grpc.fedn_pb2 as fedn
import fedn.network.grpc.fedn_pb2_grpc as rpc
from fedn.network.combiner.roundhandler import RoundConfig


class CombinerUnavailableError(Exception):
    pass


class Channel:
    """Wrapper for a gRPC channel.

    :param address: The address for the gRPC server.
    :type address: str
    :param port: The port for connecting to the gRPC server.
    :type port: int
    :param certificate: The certificate for connecting to the gRPC server (optional)
    :type certificate: str
    """

    def __init__(self, address, port, certificate=None):
        """Create a channel.

        If a valid certificate is given, a secure channel is created, else insecure.

        :parameter address: The address for the gRPC server.
        :type address: str
        :parameter port: The port for connecting to the gRPC server.
        :type port: int
        :parameter certificate: The certificate for connecting to the gRPC server (optional)
        :type certificate: str
        """
        self.address = address
        self.port = port
        self.certificate = certificate

        if self.certificate:
            credentials = grpc.ssl_channel_credentials(root_certificates=copy.deepcopy(certificate))
            self.channel = grpc.secure_channel("{}:{}".format(self.address, str(self.port)), credentials)
        else:
            self.channel = grpc.insecure_channel("{}:{}".format(self.address, str(self.port)))

    def get_channel(self):
        """Get a channel.

        :return: An instance of a gRPC channel
        :rtype: :class:`grpc.Channel`
        """
        return copy.copy(self.channel)


class CombinerInterface:
    """Interface for the Combiner (aggregation server).
        Abstraction on top of the gRPC server servicer.

    :param parent: The parent combiner (controller)
    :type parent: :class:`fedn.network.api.interfaces.API`
    :param name: The name of the combiner.
    :type name: str
    :param address: The address of the combiner.
    :type address: str
    :param fqdn: The fully qualified domain name of the combiner.
    :type fqdn: str
    :param port: The port of the combiner.
    :type port: int
    :param certificate: The certificate of the combiner (optional).
    :type certificate: str
    :param key: The key of the combiner (optional).
    :type key: str
    :param ip: The ip of the combiner (optional).
    :type ip: str
    :param config: The configuration of the combiner (optional).
    :type config: dict
    """

    def __init__(self, parent, name, address, fqdn, port, certificate=None, key=None, ip=None, config=None):
        """Initialize the combiner interface."""
        self.parent = parent
        self.name = name
        self.address = address
        self.fqdn = fqdn
        self.port = port
        self.certificate = certificate
        self.key = key
        self.ip = ip

        if not config:
            self.config = {"max_clients": 8}
        else:
            self.config = config

    @classmethod
    def from_json(combiner_config):
        """Initialize the combiner config from a json document.

        :parameter combiner_config: The combiner configuration.
        :type combiner_config: dict
        :return: An instance of the combiner interface.
        :rtype: :class:`fedn.network.combiner.interfaces.CombinerInterface`
        """
        return CombinerInterface(**combiner_config)

    def to_dict(self):
        """Export combiner configuration to a dictionary.

        :return: A dictionary with the combiner configuration.
        :rtype: dict
        """
        data = {
            "parent": self.parent,
            "name": self.name,
            "address": self.address,
            "fqdn": self.fqdn,
            "port": self.port,
            "ip": self.ip,
            "certificate": None,
            "key": None,
            "config": self.config,
        }

        if self.certificate:
            cert_b64 = base64.b64encode(self.certificate)
            key_b64 = base64.b64encode(self.key)
            data["certificate"] = str(cert_b64).split("'")[1]
            data["key"] = str(key_b64).split("'")[1]

        return data

    def to_json(self):
        """Export combiner configuration to json.

        :return: A json document with the combiner configuration.
        :rtype: str
        """
        return json.dumps(self.to_dict())

    def get_certificate(self):
        """Get combiner certificate.

        :return: The combiner certificate.
        :rtype: str, None if no certificate is set.
        """
        if self.certificate:
            cert_b64 = base64.b64encode(self.certificate)
            return str(cert_b64).split("'")[1]
        else:
            return None

    def get_key(self):
        """Get combiner key.

        :return: The combiner key.
        :rtype: str, None if no key is set.
        """
        if self.key:
            key_b64 = base64.b64encode(self.key)
            return str(key_b64).split("'")[1]
        else:
            return None

    def flush_model_update_queue(self):
        """Reset the model update queue on the combiner."""
        channel = Channel(self.address, self.port, self.certificate).get_channel()
        control = rpc.ControlStub(channel)

        request = fedn.ControlRequest()

        try:
            control.FlushAggregationQueue(request)
        except grpc.RpcError as e:
            if e.code() == grpc.StatusCode.UNAVAILABLE:
                raise CombinerUnavailableError
            else:
                raise

    def set_aggregator(self, aggregator):
        """Set the active aggregator module.

        :param aggregator: The name of the aggregator module.
        :type config: str
        """
        channel = Channel(self.address, self.port, self.certificate).get_channel()
        control = rpc.ControlStub(channel)

        request = fedn.ControlRequest()
        p = request.parameter.add()
        p.key = "aggregator"
        p.value = aggregator

        try:
            control.SetAggregator(request)
        except grpc.RpcError as e:
            if e.code() == grpc.StatusCode.UNAVAILABLE:
                raise CombinerUnavailableError
            else:
                raise

    def submit(self, config: RoundConfig):
        """Submit a compute plan to the combiner.

        :param config: The job configuration.
        :type config: dict
        :return: Server ControlResponse object.
        :rtype: :class:`fedn.network.grpc.fedn_pb2.ControlResponse`
        """
        channel = Channel(self.address, self.port, self.certificate).get_channel()
        control = rpc.ControlStub(channel)
        request = fedn.ControlRequest()
        request.command = fedn.Command.START
        for k, v in config.items():
            p = request.parameter.add()
            p.key = str(k)
            p.value = str(v)

        try:
            response = control.Start(request)
        except grpc.RpcError as e:
            if e.code() == grpc.StatusCode.UNAVAILABLE:
                raise CombinerUnavailableError
            else:
                raise

        return response

    def get_model(self, id, timeout=10):
        """Download a model from the combiner server.

        :param id: The model id.
        :type id: str
        :return: A file-like object containing the model.
        :rtype: :class:`io.BytesIO`, None if the model is not available.
        """
        channel = Channel(self.address, self.port, self.certificate).get_channel()
        modelservice = rpc.ModelServiceStub(channel)

        data = BytesIO()
        data.seek(0, 0)

        time_start = time.time()

        request = fedn.ModelRequest(id=id)
        request.sender.name = self.name
        request.sender.role = fedn.WORKER

        parts = modelservice.Download(request)
        for part in parts:
            if part.status == fedn.ModelStatus.IN_PROGRESS:
                data.write(part.data)
            if part.status == fedn.ModelStatus.OK:
                return data
            if part.status == fedn.ModelStatus.FAILED:
                return None
            if part.status == fedn.ModelStatus.UNKNOWN:
                if time.time() - time_start > timeout:
                    return None
                continue

    def allowing_clients(self):
        """Check if the combiner is allowing additional client connections.

        :return: True if accepting, else False.
        :rtype: bool
        """
        channel = Channel(self.address, self.port, self.certificate).get_channel()
        connector = rpc.ConnectorStub(channel)
        request = fedn.ConnectionRequest()

        try:
            response = connector.AcceptingClients(request)
        except grpc.RpcError as e:
            if e.code() == grpc.StatusCode.UNAVAILABLE:
                raise CombinerUnavailableError
            else:
                raise
        if response.status == fedn.ConnectionStatus.NOT_ACCEPTING:
            return False
        if response.status == fedn.ConnectionStatus.ACCEPTING:
            return True
        if response.status == fedn.ConnectionStatus.TRY_AGAIN_LATER:
            return False

        return False

    def list_active_clients(self, queue=1):
        """List active clients.

        :param queue: The channel (queue) to use (optional). Default is 1 = MODEL_UPDATE_REQUESTS channel.
            see :class:`fedn.network.grpc.fedn_pb2.Channel`
        :type channel: int
        :return: A list of active clients.
        :rtype: json
        """
        channel = Channel(self.address, self.port, self.certificate).get_channel()
        control = rpc.ConnectorStub(channel)
        request = fedn.ListClientsRequest()
        request.channel = queue
        try:
            response = control.ListActiveClients(request)
        except grpc.RpcError as e:
            if e.code() == grpc.StatusCode.UNAVAILABLE:
                raise CombinerUnavailableError
            else:
                raise
        return response.client



from flask import Blueprint, jsonify, request

from fedn.network.api.auth import jwt_auth_required
from fedn.network.api.v1.shared import api_version, get_post_data_to_kwargs, get_typed_list_headers, get_use_typing, mdb
from fedn.network.storage.statestore.stores.shared import EntityNotFound
from fedn.network.storage.statestore.stores.validation_store import ValidationStore

bp = Blueprint("validation", __name__, url_prefix=f"/api/{api_version}/validations")

validation_store = ValidationStore(mdb, "control.validations")


@bp.route("/", methods=["GET"])
@jwt_auth_required(role="admin")
def get_validations():
    """Get validations
    Retrieves a list of validations based on the provided parameters.
    By specifying a parameter in the url, you can filter the validations based on that parameter,
    and the response will contain only the validations that match the filter.
    ---
    tags:
        - Validations
    parameters:
      - name: sender.name
        in: query
        required: false
        type: string
        description: Name of the sender
      - name: sender.role
        in: query
        required: false
        type: string
        description: Role of the sender
      - name: receiver.name
        in: query
        required: false
        type: string
        description: Name of the receiver
      - name: receiver.role
        in: query
        required: false
        type: string
        description: Role of the receiver
      - name: session_id
        in: query
        required: false
        type: string
      - name: model_id
        in: query
        required: false
        type: string
      - name: correlation_id
        in: query
        required: false
        type: string
        description: Correlation id of the validation
      - name: X-Limit
        in: header
        required: false
        type: integer
        description: The maximum number of validations to retrieve
      - name: X-Skip
        in: header
        required: false
        type: integer
        description: The number of validations to skip
      - name: X-Sort-Key
        in: header
        required: false
        type: string
        description: The key to sort the validations by
      - name: X-Sort-Order
        in: header
        required: false
        type: string
        description: The order to sort the validations in ('asc' or 'desc')
    definitions:
      Validation:
        type: object
        properties:
          id:
            type: string
          correlation_id:
            type: string
          session_id:
            type: string
          model_id:
            type: string
          timestamp:
            type: object
            format: date-time
          data:
            type: string
          meta:
            type: string
          sender:
            type: object
            properties:
                name:
                    type: string
                role:
                    type: string
          receiver:
            type: object
            properties:
                name:
                    type: string
                role:
                    type: string
    responses:
      200:
        description: A list of validations and the total count.
        schema:
            type: object
            properties:
                count:
                    type: integer
                result:
                    type: array
                    items:
                        $ref: '#/definitions/Session'
      500:
        description: An error occurred
        schema:
            type: object
            properties:
                message:
                    type: string
    """
    try:
        limit, skip, sort_key, sort_order, use_typing = get_typed_list_headers(request.headers)
        kwargs = request.args.to_dict()

        validations = validation_store.list(limit, skip, sort_key, sort_order, use_typing=use_typing, **kwargs)

        result = [validation.__dict__ for validation in validations["result"]] if use_typing else validations["result"]

        response = {"count": validations["count"], "result": result}

        return jsonify(response), 200
    except Exception:
        return jsonify({"message": "An unexpected error occurred"}), 500


@bp.route("/list", methods=["POST"])
@jwt_auth_required(role="admin")
def list_validations():
    """Get validations
    Retrieves a list of validations based on the provided parameters.
    Works much like the GET validations method, but allows for a more complex query.
    By specifying a parameter in the body, you can filter the validations based on that parameter,
    and the response will contain only the validations that match the filter. If the parameter value contains a comma,
    the filter will be an "in" query, meaning that the validations will be returned if the specified field contains any of the values in the parameter.
    ---
    tags:
        - Validations
    parameters:
      - name: validation
        in: body
        required: false
        schema:
            type: object
            properties:
                sender.name:
                    type: string
                    description: Name of the sender
                sender.role:
                    required: false
                    type: string
                    description: Role of the sender
                receiver.name:
                    type: string
                    description: Name of the receiver
                receiver.role:
                    required: false
                    type: string
                    description: Role of the receiver
                session_id:
                    required: false
                    type: string
                model_id:
                    required: false
                    type: string
                correlation_id:
                    required: false
                    type: string
                    description: Correlation id of the status
      - name: X-Limit
        in: header
        required: false
        type: integer
        description: The maximum number of validations to retrieve
      - name: X-Skip
        in: header
        required: false
        type: integer
        description: The number of validations to skip
      - name: X-Sort-Key
        in: header
        required: false
        type: string
        description: The key to sort the validations by
      - name: X-Sort-Order
        in: header
        required: false
        type: string
        description: The order to sort the validations in ('asc' or 'desc')
    responses:
      200:
        description: A list of validations and the total count.
        schema:
            type: object
            properties:
                count:
                    type: integer
                result:
                    type: array
                    items:
                        $ref: '#/definitions/Session'
      500:
        description: An error occurred
        schema:
            type: object
            properties:
                message:
                    type: string
    """
    try:
        limit, skip, sort_key, sort_order, use_typing = get_typed_list_headers(request.headers)
        kwargs = get_post_data_to_kwargs(request)

        validations = validation_store.list(limit, skip, sort_key, sort_order, use_typing=use_typing, **kwargs)

        result = [validation.__dict__ for validation in validations["result"]] if use_typing else validations["result"]

        response = {"count": validations["count"], "result": result}

        return jsonify(response), 200
    except Exception:
        return jsonify({"message": "An unexpected error occurred"}), 500


@bp.route("/count", methods=["GET"])
@jwt_auth_required(role="admin")
def get_validations_count():
    """Validations count
    Retrieves the count of validations based on the provided parameters.
    By specifying a parameter in the url, you can filter the validations based on that parameter,
    and the response will contain only the count of validations that match the filter.
    ---
    tags:
        - Validations
    parameters:
      - name: sender.name
        in: query
        required: false
        type: string
        description: Name of the sender
      - name: sender.role
        in: query
        required: false
        type: string
        description: Role of the sender
      - name: receiver.name
        in: query
        required: false
        type: string
        description: Name of the receiver
      - name: receiver.role
        in: query
        required: false
        type: string
        description: Role of the receiver
      - name: session_id
        in: query
        required: false
        type: string
      - name: model_id
        in: query
        required: false
        type: string
      - name: correlation_id
        in: query
        required: false
        type: string
        description: Correlation id of the validation
    responses:
        200:
            description: The count of validations
            schema:
                type: integer
        500:
            description: An error occurred
            schema:
                type: object
                properties:
                    message:
                        type: string
    """
    try:
        kwargs = request.args.to_dict()
        count = validation_store.count(**kwargs)
        response = count
        return jsonify(response), 200
    except Exception:
        return jsonify({"message": "An unexpected error occurred"}), 500


@bp.route("/count", methods=["POST"])
@jwt_auth_required(role="admin")
def validations_count():
    """Validations count
    Retrieves the count of validations based on the provided parameters.
    Works much like the GET /validations/count endpoint, but allows for more complex queries.
    By specifying a parameter in the request body, you can filter the validations based on that parameter,
    if the parameter value contains a comma, the filter will be an "in" query, meaning that the validations
    will be returned if the specified field contains any of the values in the parameter.
    ---
    tags:
        - Validations
    parameters:
      - name: validation
        in: body
        required: false
        schema:
            type: object
            properties:
                sender.name:
                    type: string
                    description: Name of the sender
                sender.role:
                    required: false
                    type: string
                    description: Role of the sender
                receiver.name:
                    type: string
                    description: Name of the receiver
                receiver.role:
                    required: false
                    type: string
                    description: Role of the receiver
                session_id:
                    required: false
                    type: string
                model_id:
                    required: false
                    type: string
                correlation_id:
                    required: false
                    type: string
                    description: Correlation id of the status
    responses:
        200:
            description: The count of validations
            schema:
                type: integer
        500:
            description: An error occurred
            schema:
                type: object
                properties:
                    message:
                        type: string
    """
    try:
        kwargs = get_post_data_to_kwargs(request)
        count = validation_store.count(**kwargs)
        response = count
        return jsonify(response), 200
    except Exception:
        return jsonify({"message": "An unexpected error occurred"}), 500


@bp.route("/<string:id>", methods=["GET"])
@jwt_auth_required(role="admin")
def get_validation(id: str):
    """Get validation
    Retrieves a validation based on the provided id.
    ---
    tags:
        - Validations
    parameters:
      - name: id
        in: path
        required: true
        type: string
        description: The id of the validation
    responses:
        200:
            description: The validation
            schema:
                $ref: '#/definitions/Validation'
        404:
            description: The validation was not found
            schema:
                type: object
                properties:
                    error:
                        type: string
        500:
            description: An error occurred
            schema:
                type: object
                properties:
                    message:
                        type: string
    """
    try:
        use_typing: bool = get_use_typing(request.headers)
        validation = validation_store.get(id, use_typing=use_typing)

        response = validation.__dict__ if use_typing else validation

        return jsonify(response), 200
    except EntityNotFound:
        return jsonify({"message": f"Entity with id: {id} not found"}), 404
    except Exception:
        return jsonify({"message": "An unexpected error occurred"}), 500



from functools import wraps

import jwt
from flask import jsonify, request

from fedn.common.config import (
    FEDN_AUTH_SCHEME,
    FEDN_AUTH_WHITELIST_URL_PREFIX,
    FEDN_JWT_ALGORITHM,
    FEDN_JWT_CUSTOM_CLAIM_KEY,
    FEDN_JWT_CUSTOM_CLAIM_VALUE,
    SECRET_KEY,
)


def check_role_claims(payload, role):
    if "role" not in payload:
        return False
    if payload["role"] != role:
        return False

    return True


def check_custom_claims(payload):
    if FEDN_JWT_CUSTOM_CLAIM_KEY and FEDN_JWT_CUSTOM_CLAIM_VALUE:
        if payload[FEDN_JWT_CUSTOM_CLAIM_KEY] != FEDN_JWT_CUSTOM_CLAIM_VALUE:
            return False
    return True


def if_whitelisted_url_prefix(path):
    if FEDN_AUTH_WHITELIST_URL_PREFIX and path.startswith(FEDN_AUTH_WHITELIST_URL_PREFIX):
        return True
    else:
        return False


def jwt_auth_required(role=None):
    def actual_decorator(func):
        if not SECRET_KEY:
            return func

        @wraps(func)
        def decorated(*args, **kwargs):
            if if_whitelisted_url_prefix(request.path):
                return func(*args, **kwargs)
            token = request.headers.get("Authorization")
            if not token:
                return jsonify({"message": "Missing token"}), 401
            # Get token from the header Bearer
            if token.startswith(FEDN_AUTH_SCHEME):
                token = token.split(" ")[1]
            else:
                return jsonify({"message": f"Invalid token scheme, expected {FEDN_AUTH_SCHEME}"}), 401
            try:
                payload = jwt.decode(token, SECRET_KEY, algorithms=[FEDN_JWT_ALGORITHM])
                if not check_role_claims(payload, role):
                    return jsonify({"message": "Invalid token"}), 401
                if not check_custom_claims(payload):
                    return jsonify({"message": "Invalid token"}), 401

            except jwt.ExpiredSignatureError:
                return jsonify({"message": "Token expired"}), 401

            except jwt.InvalidTokenError:
                return jsonify({"message": "Invalid token"}), 401

            return func(*args, **kwargs)

        return decorated

    return actual_decorator



import unittest
from unittest.mock import MagicMock

from fedn.network.combiner.aggregators.fedavg import FedAvg


class TestFedAvg(unittest.TestCase):
    """Test the FedAvg aggregator"""

    def SetUp(self):
        pass

    def test_fedavg_init(self, *args, **kwargs):
        """Test the FedAvg aggregator constructor"""
        aggregator = FedAvg("id", None, None, None, None)
        self.assertEqual(aggregator.name, "FedAvg")

    def test_fedavg_combine_models(self, *args, **kwargs):
        """Test the FedAvg aggregator combine_models method with mock classes and methods"""
        aggregator = FedAvg("id", None, None, None, None)
        aggregator.next_model_update = MagicMock(return_value=(None, None, None))
        aggregator.server = MagicMock()

        data = {}
        data['time_model_load'] = 0.0
        data['time_model_aggregation'] = 0.0
        data['nr_aggregated_models'] = 0

        self.assertEqual(aggregator.combine_models(), (None, data))


if __name__ == '__main__':
    unittest.main()



""" The utils package is responsible for providing utility functions for the FEDn framework. Such as logging, checksums and other model helper functions to aggregate models.
THe helper functions is there to support aggregating various models from different ML frameworks, such as Tensorflow, PyTorch and Keras."""
# flake8: noqa
from sys import version_info

PYTHON_VERSION = f"{version_info.major}.{version_info.minor}.{version_info.micro}"



import hashlib


def sha(fname):
    """Calculate the sha256 checksum of a file. Used for computing checksums of compute packages.

    :param fname: The file path.
    :type fname: str
    :return: The sha256 checksum.
    :rtype: :py:class:`hashlib.sha256`
    """
    hash = hashlib.sha256()
    with open(fname, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            hash.update(chunk)
    return hash.hexdigest()



from fedn.utils.helpers.plugins.numpyhelper import Helper


class Helper(Helper):
    """FEDn helper class for models weights/parameters that can be transformed to numpy ndarrays."""

    def __init__(self):
        """Initialize helper."""
        super().__init__()
        self.name = "binaryhelper"

    def load(self, path, file_type="raw_binary"):
        return super().load(path, file_type)

    def save(self, model, path=None, file_type="raw_binary"):
        return super().save(model, path, file_type)



from flask import Blueprint, jsonify, request

from fedn.network.api.auth import jwt_auth_required
from fedn.network.api.v1.shared import api_version, get_post_data_to_kwargs, get_typed_list_headers, mdb
from fedn.network.storage.statestore.stores.client_store import ClientStore
from fedn.network.storage.statestore.stores.shared import EntityNotFound

bp = Blueprint("client", __name__, url_prefix=f"/api/{api_version}/clients")

client_store = ClientStore(mdb, "network.clients")


@bp.route("/", methods=["GET"])
@jwt_auth_required(role="admin")
def get_clients():
    """Get clients
    Retrieves a list of clients based on the provided parameters.
    By specifying a parameter in the url, you can filter the clients based on that parameter,
    and the response will contain only the clients that match the filter.
    ---
    tags:
        - Clients
    parameters:
      - name: name
        in: query
        required: false
        type: string
        description: The name of the client
      - name: combiner
        in: query
        required: false
        type: string
        description: The combiner (id) that the client has connected to
      - name: combiner_preferred
        in: query
        required: false
        type: string
        description: The combiner (id) that the client has preferred to connect to
      - name: ip
        in: query
        required: false
        type: string
        description: The ip of the client
      - name: status
        in: query
        required: false
        type: string
        description: The status of the client
      - name: updated_at
        in: query
        required: false
        type: string
        description: The date and time the client was last updated
      - name: X-Limit
        in: header
        required: false
        type: integer
        description: The maximum number of clients to retrieve
      - name: X-Skip
        in: header
        required: false
        type: integer
        description: The number of clients to skip
      - name: X-Sort-Key
        in: header
        required: false
        type: string
        description: The key to sort the clients by
      - name: X-Sort-Order
        in: header
        required: false
        type: string
        description: The order to sort the clients in ('asc' or 'desc')
    definitions:
      Client:
        type: object
        properties:
          name:
            type: string
          combiner:
            type: string
          combiner_preferred:
            type: string
            description: The combiner (id) that the client has preferred to connect or false (boolean) if the client has no preferred combiner
          ip:
            type: string
          status:
            type: string
          updated_at:
            type: string
          last_seen:
            type: string
    responses:
      200:
        description: A list of clients and the total count.
        schema:
            type: object
            properties:
                count:
                    type: integer
                result:
                    type: array
                    items:
                        $ref: '#/definitions/Client'
      500:
        description: An error occurred
        schema:
            type: object
            properties:
                message:
                    type: string
    """
    try:
        limit, skip, sort_key, sort_order, _ = get_typed_list_headers(request.headers)
        kwargs = request.args.to_dict()

        clients = client_store.list(limit, skip, sort_key, sort_order, use_typing=False, **kwargs)

        result = clients["result"]

        response = {"count": clients["count"], "result": result}

        return jsonify(response), 200
    except Exception:
        return jsonify({"message": "An unexpected error occurred"}), 500


@bp.route("/list", methods=["POST"])
@jwt_auth_required(role="admin")
def list_clients():
    """List clients
    Retrieves a list of clients based on the provided parameters.
    Works much like the GET /clients endpoint, but allows for more complex queries.
    By specifying a parameter in the request body, you can filter the clients based on that parameter,
    and the response will contain only the clients that match the filter. If the parameter value contains a comma,
    the filter will be an "in" query, meaning that the clients will be returned if the specified field contains any of the values in the parameter.
    ---
    tags:
        - Clients
    parameters:
      - name: client
        in: body
        required: false
        type: object
        description: Object containing the parameters to filter the clients by
        schema:
          type: object
          properties:
            name:
              type: string
            combiner:
              type: string
            ip:
              type: string
            status:
              type: string
            updated_at:
              type: string
      - name: X-Limit
        in: header
        required: false
        type: integer
        description: The maximum number of clients to retrieve
      - name: X-Skip
        in: header
        required: false
        type: integer
        description: The number of clients to skip
      - name: X-Sort-Key
        in: header
        required: false
        type: string
        description: The key to sort the clients by
      - name: X-Sort-Order
        in: header
        required: false
        type: string
        description: The order to sort the clients in ('asc' or 'desc')
    responses:
      200:
        description: A list of clients and the total count.
        schema:
            type: object
            properties:
                count:
                    type: integer
                result:
                    type: array
                    items:
                        $ref: '#/definitions/Client'
      500:
        description: An error occurred
        schema:
            type: object
            properties:
                message:
                    type: string
    """
    try:
        limit, skip, sort_key, sort_order, _ = get_typed_list_headers(request.headers)
        kwargs = get_post_data_to_kwargs(request)
        clients = client_store.list(limit, skip, sort_key, sort_order, use_typing=False, **kwargs)

        result = clients["result"]

        response = {"count": clients["count"], "result": result}

        return jsonify(response), 200
    except Exception:
        return jsonify({"message": "An unexpected error occurred"}), 500


@bp.route("/count", methods=["GET"])
@jwt_auth_required(role="admin")
def get_clients_count():
    """Clients count
    Retrieves the total number of clients based on the provided parameters.
    ---
    tags:
        - Clients
    parameters:
      - name: name
        in: query
        required: false
        type: string
        description: The name of the client
      - name: combiner
        in: query
        required: false
        type: string
        description: The combiner (id) that the client has connected to
      - name: combiner_preferred
        in: query
        required: false
        type: string
        description: The combiner (id) that the client has preferred to connect to
      - name: ip
        in: query
        required: false
        type: string
        description: The ip of the client
      - name: status
        in: query
        required: false
        type: string
        description: The status of the client
      - name: updated_at
        in: query
        required: false
        type: string
        description: The date and time the client was last updated
    responses:
      200:
        description: A list of clients and the total count.
        schema:
            type: integer
      500:
        description: An error occurred
        schema:
            type: object
            properties:
                message:
                    type: string
    """
    try:
        kwargs = request.args.to_dict()
        count = client_store.count(**kwargs)
        response = count
        return jsonify(response), 200
    except Exception:
        return jsonify({"message": "An unexpected error occurred"}), 500


@bp.route("/count", methods=["POST"])
@jwt_auth_required(role="admin")
def clients_count():
    """Clients count
    Retrieves the total number of clients based on the provided parameters.
    Works much like the GET /clients/count endpoint, but allows for more complex queries.
    By specifying a parameter in the request body, you can filter the clients based on that parameter,
    and the response will contain only the clients that match the filter. If the parameter value contains a comma,
    the filter will be an "in" query, meaning that the clients will be returned if the specified field contains any of the values in the parameter.
    ---
    tags:
        - Clients
    parameters:
      - name: client
        in: body
        required: false
        type: object
        description: Object containing the parameters to filter the clients by
        schema:
          type: object
          properties:
            name:
              type: string
            combiner:
              type: string
            ip:
              type: string
            status:
              type: string
            updated_at:
              type: string
    responses:
      200:
        description: A list of clients and the total count.
        schema:
            type: integer
      500:
        description: An error occurred
        schema:
            type: object
            properties:
                message:
                    type: string
    """
    try:
        kwargs = get_post_data_to_kwargs(request)
        count = client_store.count(**kwargs)
        response = count
        return jsonify(response), 200
    except Exception:
        return jsonify({"message": "An unexpected error occurred"}), 500


@bp.route("/<string:id>", methods=["GET"])
@jwt_auth_required(role="admin")
def get_client(id: str):
    """Get client
    Retrieves a client based on the provided id.
    ---
    tags:
        - Clients
    parameters:
      - name: id
        in: path
        required: true
        type: string
        description: The id of the client
    responses:
        200:
            description: A client object
            schema:
                $ref: '#/definitions/Client'
        404:
            description: The client was not found
            schema:
                type: object
                properties:
                    message:
                        type: string
        500:
            description: An error occurred
            schema:
                type: object
                properties:
                    message:
                        type: string
    """
    try:
        client = client_store.get(id, use_typing=False)

        response = client

        return jsonify(response), 200
    except EntityNotFound:
        return jsonify({"message": f"Entity with id: {id} not found"}), 404
    except Exception:
        return jsonify({"message": "An unexpected error occurred"}), 500



import os
import tarfile

import click
import requests

from fedn.common.log_config import logger

from .main import main
from .shared import CONTROLLER_DEFAULTS, get_api_url, get_token, print_response


@main.group("package")
@click.pass_context
def package_cmd(ctx):
    """:param ctx:
    """
    pass


@package_cmd.command("create")
@click.option("-p", "--path", required=True, help="Path to package directory containing fedn.yaml")
@click.option("-n", "--name", required=False, default="package.tgz", help="Name of package tarball")
@click.pass_context
def create_cmd(ctx, path, name):
    """Create compute package.

    Make a tar.gz archive of folder given by --path

    :param ctx:
    :param path:
    """
    path = os.path.abspath(path)
    yaml_file = os.path.join(path, "fedn.yaml")
    if not os.path.exists(yaml_file):
        logger.error(f"Could not find fedn.yaml in {path}")
        exit(-1)

    with tarfile.open(name, "w:gz") as tar:
        tar.add(path, arcname=os.path.basename(path))
        logger.info(f"Created package {name}")


@click.option("-p", "--protocol", required=False, default=CONTROLLER_DEFAULTS["protocol"], help="Communication protocol of controller (api)")
@click.option("-H", "--host", required=False, default=CONTROLLER_DEFAULTS["host"], help="Hostname of controller (api)")
@click.option("-P", "--port", required=False, default=CONTROLLER_DEFAULTS["port"], help="Port of controller (api)")
@click.option("-t", "--token", required=False, help="Authentication token")
@click.option("--n_max", required=False, help="Number of items to list")
@package_cmd.command("list")
@click.pass_context
def list_packages(ctx, protocol: str, host: str, port: str, token: str = None, n_max: int = None):
    """Return:
    ------
    - count: number of packages
    - result: list of packages

    """
    url = get_api_url(protocol=protocol, host=host, port=port, endpoint="packages")
    headers = {}

    if n_max:
        headers["X-Limit"] = n_max

    _token = get_token(token)

    if _token:
        headers["Authorization"] = _token

    click.echo(f"\nListing packages: {url}\n")
    click.echo(f"Headers: {headers}")

    try:
        response = requests.get(url, headers=headers)
        print_response(response, "packages")
    except requests.exceptions.ConnectionError:
        click.echo(f"Error: Could not connect to {url}")



import ast
import queue
import random
import sys
import time
import uuid
from typing import TypedDict

from fedn.common.log_config import logger
from fedn.network.combiner.aggregators.aggregatorbase import get_aggregator
from fedn.network.combiner.modelservice import load_model_from_BytesIO, serialize_model_to_BytesIO
from fedn.utils.helpers.helpers import get_helper
from fedn.utils.parameters import Parameters


class RoundConfig(TypedDict):
    """Round configuration.

    :param _job_id: A universally unique identifier for the round. Set by Combiner.
    :type _job_id: str
    :param committed_at: The time the round was committed. Set by Controller.
    :type committed_at: str
    :param task: The task to perform in the round. Set by Controller. Supported tasks are "training", "validation", and "inference".
    :type task: str
    :param round_id: The round identifier as str(int)
    :type round_id: str
    :param round_timeout: The round timeout in seconds. Set by user interfaces or Controller.
    :type round_timeout: str
    :param rounds: The number of rounds. Set by user interfaces.
    :param model_id: The model identifier. Set by user interfaces or Controller (get_latest_model).
    :type model_id: str
    :param model_version: The model version. Currently not used.
    :type model_version: str
    :param model_type: The model type. Currently not used.
    :type model_type: str
    :param model_size: The size of the model. Currently not used.
    :type model_size: int
    :param model_parameters: The model parameters. Currently not used.
    :type model_parameters: dict
    :param model_metadata: The model metadata. Currently not used.
    :type model_metadata: dict
    :param session_id: The session identifier. Set by (Controller?).
    :type session_id: str
    :param helper_type: The helper type.
    :type helper_type: str
    :param aggregator: The aggregator type.
    :type aggregator: str
    """

    _job_id: str
    committed_at: str
    task: str
    round_id: str
    round_timeout: str
    rounds: int
    model_id: str
    model_version: str
    model_type: str
    model_size: int
    model_parameters: dict
    model_metadata: dict
    session_id: str
    helper_type: str
    aggregator: str


class ModelUpdateError(Exception):
    pass


class RoundHandler:
    """Round handler.

    The round handler processes requests from the global controller
    to produce model updates and perform model validations.

    :param aggregator_name: The name of the aggregator plugin module.
    :type aggregator_name: str
    :param storage: Model repository for :class: `fedn.network.combiner.Combiner`
    :type storage: class: `fedn.common.storage.s3.s3repo.S3ModelRepository`
    :param server: A handle to the Combiner class :class: `fedn.network.combiner.Combiner`
    :type server: class: `fedn.network.combiner.Combiner`
    :param modelservice: A handle to the model service :class: `fedn.network.combiner.modelservice.ModelService`
    :type modelservice: class: `fedn.network.combiner.modelservice.ModelService`
    """

    def __init__(self, storage, server, modelservice):
        """Initialize the RoundHandler."""
        self.round_configs = queue.Queue()
        self.storage = storage
        self.server = server
        self.modelservice = modelservice

    def set_aggregator(self, aggregator):
        self.aggregator = get_aggregator(aggregator, self.storage, self.server, self.modelservice, self)

    def push_round_config(self, round_config: RoundConfig) -> str:
        """Add a round_config (job description) to the inbox.

        :param round_config: A dict containing the round configuration (from global controller).
        :type round_config: dict
        :return: A job id (universally unique identifier) for the round.
        :rtype: str
        """
        try:
            round_config["_job_id"] = str(uuid.uuid4())
            self.round_configs.put(round_config)
        except Exception:
            logger.error("Failed to push round config.")
            raise
        return round_config["_job_id"]

    def load_model_update(self, helper, model_id):
        """Load model update with id model_id into its memory representation.

        :param helper: An instance of :class: `fedn.utils.helpers.helpers.HelperBase`
        :type helper: class: `fedn.utils.helpers.helpers.HelperBase`
        :param model_id: The ID of the model update, UUID in str format
        :type model_id: str
        """
        model_str = self.load_model_update_str(model_id)
        if model_str:
            try:
                model = load_model_from_BytesIO(model_str.getbuffer(), helper)
            except IOError:
                logger.warning("AGGREGATOR({}): Failed to load model!".format(self.name))
        else:
            raise ModelUpdateError("Failed to load model.")

        return model

    def load_model_update_str(self, model_id, retry=3):
        """Load model update object and return it as BytesIO.

        :param model_id: The ID of the model
        :type model_id: str
        :param retry: number of times retrying load model update, defaults to 3
        :type retry: int, optional
        :return: Updated model
        :rtype: class: `io.BytesIO`
        """
        # Try reading model update from local disk/combiner memory
        model_str = self.modelservice.temp_model_storage.get(model_id)
        # And if we cannot access that, try downloading from the server
        if model_str is None:
            model_str = self.modelservice.get_model(model_id)
            # TODO: use retrying library
            tries = 0
            while tries < retry:
                tries += 1
                if not model_str or sys.getsizeof(model_str) == 80:
                    logger.warning("Model download failed. retrying")
                    time.sleep(1)
                    model_str = self.modelservice.get_model(model_id)

        return model_str

    def waitforit(self, config, buffer_size=100, polling_interval=0.1):
        """Defines the policy for how long the server should wait before starting to aggregate models.

        The policy is as follows:
            1. Wait a maximum of time_window time until the round times out.
            2. Terminate if a preset number of model updates (buffer_size) are in the queue.

        :param config: The round config object
        :type config: dict
        :param buffer_size: The number of model updates to wait for before starting aggregation, defaults to 100
        :type buffer_size: int, optional
        :param polling_interval: The polling interval, defaults to 0.1
        :type polling_interval: float, optional
        """
        time_window = float(config["round_timeout"])

        tt = 0.0
        while tt < time_window:
            if self.aggregator.model_updates.qsize() >= buffer_size:
                break

            time.sleep(polling_interval)
            tt += polling_interval

    def _training_round(self, config, clients):
        """Send model update requests to clients and aggregate results.

        :param config: The round config object (passed to the client).
        :type config: dict
        :param clients: clients to participate in the training round
        :type clients: list
        :return: an aggregated model and associated metadata
        :rtype: model, dict
        """
        logger.info("ROUNDHANDLER: Initiating training round, participating clients: {}".format(clients))

        meta = {}
        meta["nr_expected_updates"] = len(clients)
        meta["nr_required_updates"] = int(config["clients_required"])
        meta["timeout"] = float(config["round_timeout"])

        session_id = config["session_id"]
        model_id = config["model_id"]

        # Request model updates from all active clients.
        self.server.request_model_update(session_id=session_id, model_id=model_id, config=config, clients=clients)

        # If buffer_size is -1 (default), the round terminates when/if all clients have completed.
        if int(config["buffer_size"]) == -1:
            buffer_size = len(clients)
        else:
            buffer_size = int(config["buffer_size"])

        # Wait / block until the round termination policy has been met.
        self.waitforit(config, buffer_size=buffer_size)

        tic = time.time()
        model = None
        data = None

        try:
            helper = get_helper(config["helper_type"])
            logger.info("Config delete_models_storage: {}".format(config["delete_models_storage"]))
            if config["delete_models_storage"] == "True":
                delete_models = True
            else:
                delete_models = False

            if "aggregator_kwargs" in config.keys():
                dict_parameters = ast.literal_eval(config["aggregator_kwargs"])
                parameters = Parameters(dict_parameters)
            else:
                parameters = None

            model, data = self.aggregator.combine_models(helper=helper, delete_models=delete_models, parameters=parameters)
        except Exception as e:
            logger.warning("AGGREGATION FAILED AT COMBINER! {}".format(e))

        meta["time_combination"] = time.time() - tic
        meta["aggregation_time"] = data
        return model, meta

    def _validation_round(self, session_id, model_id, clients):
        """Send model validation requests to clients.

        :param config: The round config object (passed to the client).
        :type config: dict
        :param clients: clients to send validation requests to
        :type clients: list
        :param model_id: The ID of the model to validate
        :type model_id: str
        """
        self.server.request_model_validation(session_id, model_id, clients=clients)

    def _inference_round(self, session_id: str, model_id: str, clients: list):
        """Send model inference requests to clients.

        :param config: The round config object (passed to the client).
        :type config: dict
        :param clients: clients to send inference requests to
        :type clients: list
        :param model_id: The ID of the model to use for inference
        :type model_id: str
        """
        self.server.request_model_inference(session_id, model_id, clients=clients)

    def stage_model(self, model_id, timeout_retry=3, retry=2):
        """Download a model from persistent storage and set in modelservice.

        :param model_id: ID of the model update object to stage.
        :type model_id: str
        :param timeout_retry: Sleep before retrying download again(sec), defaults to 3
        :type timeout_retry: int, optional
        :param retry: Number of retries, defaults to 2
        :type retry: int, optional
        """
        # If the model is already in memory at the server we do not need to do anything.
        if self.modelservice.temp_model_storage.exist(model_id):
            logger.info("Model already exists in memory, skipping model staging.")
            return
        logger.info("Model Staging, fetching model from storage...")
        # If not, download it and stage it in memory at the combiner.
        tries = 0
        while True:
            try:
                model = self.storage.get_model_stream(model_id)
                if model:
                    break
            except Exception:
                logger.warning("Could not fetch model from storage backend, retrying.")
                time.sleep(timeout_retry)
                tries += 1
                if tries > retry:
                    logger.error("Failed to stage model {} from storage backend!".format(model_id))
                    raise

        self.modelservice.set_model(model, model_id)

    def _assign_round_clients(self, n, type="trainers"):
        """Obtain a list of clients(trainers or validators) to ask for updates in this round.

        :param n: Size of a random set taken from active trainers(clients), if n > "active trainers" all is used
        :type n: int
        :param type: type of clients, either "trainers" or "validators", defaults to "trainers"
        :type type: str, optional
        :return: Set of clients
        :rtype: list
        """
        if type == "validators":
            clients = self.server.get_active_validators()
        elif type == "trainers":
            clients = self.server.get_active_trainers()
        else:
            logger.error("(ERROR): {} is not a supported type of client".format(type))

        # If the number of requested trainers exceeds the number of available, use all available.
        if n > len(clients):
            n = len(clients)

        # If not, we pick a random subsample of all available clients.
        clients = random.sample(clients, n)

        return clients

    def _check_nr_round_clients(self, config):
        """Check that the minimal number of clients required to start a round are available.

        :param config: The round config object.
        :type config: dict
        :param timeout: Timeout in seconds, defaults to 0.0
        :type timeout: float, optional
        :return: True if the required number of clients are available, False otherwise.
        :rtype: bool
        """
        active = self.server.nr_active_trainers()
        if active >= int(config["clients_required"]):
            logger.info("Number of clients required ({0}) to start round met {1}.".format(config["clients_required"], active))
            return True
        else:
            logger.info("Too few clients to start round.")
            return False

    def execute_validation_round(self, session_id, model_id):
        """Coordinate validation rounds as specified in config.

        :param round_config: The round config object.
        :type round_config: dict
        """
        logger.info("COMBINER orchestrating validation of model {}".format(model_id))
        self.stage_model(model_id)
        validators = self._assign_round_clients(self.server.max_clients, type="validators")
        self._validation_round(session_id, model_id, validators)

    def execute_inference_round(self, session_id: str, model_id: str) -> None:
        """Coordinate inference rounds as specified in config.

        :param round_config: The round config object.
        :type round_config: dict
        """
        logger.info("COMBINER orchestrating inference using model {}".format(model_id))
        self.stage_model(model_id)
        # TODO: Implement inference client type
        clients = self._assign_round_clients(self.server.max_clients, type="validators")
        self._inference_round(session_id, model_id, clients)

    def execute_training_round(self, config):
        """Coordinates clients to execute training tasks.

        :param config: The round config object.
        :type config: dict
        :return: metadata about the training round.
        :rtype: dict
        """
        logger.info("Processing training round,  job_id {}".format(config["_job_id"]))

        data = {}
        data["config"] = config
        data["round_id"] = config["round_id"]

        # Download model to update and set in temp storage.
        self.stage_model(config["model_id"])

        clients = self._assign_round_clients(self.server.max_clients)
        model, meta = self._training_round(config, clients)
        data["data"] = meta

        if model is None:
            logger.warning("\t Failed to update global model in round {0}!".format(config["round_id"]))

        if model is not None:
            helper = get_helper(config["helper_type"])
            a = serialize_model_to_BytesIO(model, helper)
            model_id = self.storage.set_model(a.read(), is_file=False)
            a.close()
            data["model_id"] = model_id

            logger.info("TRAINING ROUND COMPLETED. Aggregated model id: {}, Job id: {}".format(model_id, config["_job_id"]))

        # Delete temp model
        self.modelservice.temp_model_storage.delete(config["model_id"])
        return data

    def run(self, polling_interval=1.0):
        """Main control loop. Execute rounds based on round config on the queue.

        :param polling_interval: The polling interval in seconds for checking if a new job/config is available.
        :type polling_interval: float
        """
        try:
            while True:
                try:
                    round_config = self.round_configs.get(block=False)
                    session_id = round_config["session_id"]
                    model_id = round_config["model_id"]

                    # Check that the minimum allowed number of clients are connected
                    ready = self._check_nr_round_clients(round_config)
                    round_meta = {}

                    if ready:
                        if round_config["task"] == "training":
                            tic = time.time()
                            round_meta = self.execute_training_round(round_config)
                            round_meta["time_exec_training"] = time.time() - tic
                            round_meta["status"] = "Success"
                            round_meta["name"] = self.server.id
                            self.server.statestore.set_round_combiner_data(round_meta)
                        elif round_config["task"] == "validation":
                            self.execute_validation_round(session_id, model_id)
                        elif round_config["task"] == "inference":
                            self.execute_inference_round(session_id, model_id)
                        else:
                            logger.warning("config contains unkown task type.")
                    else:
                        round_meta = {}
                        round_meta["status"] = "Failed"
                        round_meta["reason"] = "Failed to meet client allocation requirements for this round config."
                        logger.warning("{0}".format(round_meta["reason"]))

                    self.round_configs.task_done()
                except queue.Empty:
                    time.sleep(polling_interval)

        except (KeyboardInterrupt, SystemExit):
            pass



from typing import Any, Dict, List, Tuple

import pymongo
from bson import ObjectId
from pymongo.database import Database

from fedn.network.storage.statestore.stores.store import Store

from .shared import EntityNotFound, from_document


class Combiner:
    def __init__(
            self,
            id: str,
            name: str,
            address: str,
            certificate: str,
            config: dict,
            fqdn: str,
            ip: str,
            key: str,
            parent: dict,
            port: int,
            status: str,
            updated_at: str
    ):
        self.id = id
        self.name = name
        self.address = address
        self.certificate = certificate
        self.config = config
        self.fqdn = fqdn
        self.ip = ip
        self.key = key
        self.parent = parent
        self.port = port
        self.status = status
        self.updated_at = updated_at

    def from_dict(data: dict) -> "Combiner":
        return Combiner(
            id=str(data["_id"]),
            name=data["name"] if "name" in data else None,
            address=data["address"] if "address" in data else None,
            certificate=data["certificate"] if "certificate" in data else None,
            config=data["config"] if "config" in data else None,
            fqdn=data["fqdn"] if "fqdn" in data else None,
            ip=data["ip"] if "ip" in data else None,
            key=data["key"] if "key" in data else None,
            parent=data["parent"] if "parent" in data else None,
            port=data["port"] if "port" in data else None,
            status=data["status"] if "status" in data else None,
            updated_at=data["updated_at"] if "updated_at" in data else None
        )


class CombinerStore(Store[Combiner]):
    def __init__(self, database: Database, collection: str):
        super().__init__(database, collection)

    def get(self, id: str, use_typing: bool = False) -> Combiner:
        """Get an entity by id
        param id: The id of the entity
            type: str
            description: The id of the entity, can be either the id or the name (property)
        param use_typing: Whether to return the entity as a typed object or as a dict
            type: bool
        return: The entity
        """
        if ObjectId.is_valid(id):
            id_obj = ObjectId(id)
            document = self.database[self.collection].find_one({"_id": id_obj})
        else:
            document = self.database[self.collection].find_one({"name": id})

        if document is None:
            raise EntityNotFound(f"Entity with (id | name) {id} not found")

        return Combiner.from_dict(document) if use_typing else from_document(document)

    def update(self, id: str, item: Combiner) -> bool:
        raise NotImplementedError("Update not implemented for CombinerStore")

    def add(self, item: Combiner)-> Tuple[bool, Any]:
        raise NotImplementedError("Add not implemented for CombinerStore")

    def delete(self, id: str) -> bool:
        raise NotImplementedError("Delete not implemented for CombinerStore")

    def list(self, limit: int, skip: int, sort_key: str, sort_order=pymongo.DESCENDING, use_typing: bool = False, **kwargs) -> Dict[int, List[Combiner]]:
        """List entities
        param limit: The maximum number of entities to return
            type: int
        param skip: The number of entities to skip
            type: int
        param sort_key: The key to sort by
            type: str
        param sort_order: The order to sort by
            type: pymongo.DESCENDING | pymongo.ASCENDING
        param use_typing: Whether to return the entities as typed objects or as dicts
            type: bool
        param kwargs: Additional query parameters
            type: dict
            example: {"key": "models"}
        return: A dictionary with the count and the result
        """
        response = super().list(limit, skip, sort_key or "updated_at", sort_order, use_typing=use_typing, **kwargs)

        result = [Combiner.from_dict(item) for item in response["result"]] if use_typing else response["result"]

        return {
            "count": response["count"],
            "result": result
        }

    def count(self, **kwargs) -> int:
        return super().count(**kwargs)



from flask import Blueprint, jsonify, request

from fedn.network.api.auth import jwt_auth_required
from fedn.network.api.v1.shared import api_version, get_post_data_to_kwargs, get_typed_list_headers, get_use_typing, mdb
from fedn.network.storage.statestore.stores.package_store import PackageStore
from fedn.network.storage.statestore.stores.shared import EntityNotFound

bp = Blueprint("package", __name__, url_prefix=f"/api/{api_version}/packages")

package_store = PackageStore(mdb, "control.package")


@bp.route("/", methods=["GET"])
@jwt_auth_required(role="admin")
def get_packages():
    """Get packages
    Retrieves a list of packages based on the provided parameters.
    By specifying a parameter in the url, you can filter the packages based on that parameter,
    and the response will contain only the packages that match the filter.
    ---
    tags:
      - Packages
    parameters:
      - in: query
        name: name
        schema:
        type: string
        description: The name of the package
      - in: query
        name: description
        schema:
        type: string
        description: The description of the package
      - in: query
        name: file_name
        schema:
        type: string
        description: Display name of the file, set by the user
      - in: query
        name: storage_file_name
        schema:
        type: string
        description: Name of the stored file, generated by the system
      - name: X-Limit
        in: header
        required: false
        type: integer
        description: The maximum number of packages to retrieve
      - name: X-Skip
        in: header
        required: false
        type: integer
        description: The number of packages to skip
      - name: X-Sort-Key
        in: header
        required: false
        type: string
        description: The key to sort the packages by
      - name: X-Sort-Order
        in: header
        required: false
        type: string
        description: The order to sort the packages in ('asc' or 'desc')
    definitions:
        Package:
            type: object
            properties:
            id:
                type: string
                description: The id of the package
            name:
                type: string
                description: Name of the package
            description:
                type: string
                description: Description of the package
            file_name:
                type: string
                description: Display name of the file
            storage_file_name:
                type: string
                description: Name of the stored file
            key:
                type: string
                description: The key of the type of document. Will be 'package_trail'
            helper:
                type: string
                description: The helper setting for the package
            active:
                type: boolean
                description: Whether the package is active (used if session is started) or not
            committed_at:
                type: string
                format: date-time
                description: The date and time the package was created
    responses:
      200:
        description: A list of packages and the total count.
        schema:
            type: object
            properties:
                count:
                    type: integer
                result:
                    type: array
                    items:
                        $ref: '#/definitions/Package'
      500:
        description: An error occurred
        schema:
            type: object
            properties:
                error:
                    type: string

    """
    try:
        limit, skip, sort_key, sort_order, _ = get_typed_list_headers(request.headers)
        kwargs = request.args.to_dict()

        packages = package_store.list(limit, skip, sort_key, sort_order, use_typing=True, **kwargs)

        result = [package.__dict__ for package in packages["result"]]

        response = {"count": packages["count"], "result": result}

        return jsonify(response), 200
    except Exception:
        return jsonify({"message": "An unexpected error occurred"}), 500


@bp.route("/list", methods=["POST"])
@jwt_auth_required(role="admin")
def list_packages():
    """List packages
    Retrieves a list of packages based on the provided parameters.
    Works much like the GET /packages endpoint, but allows for more complex queries.
    By specifying a parameter in the body, you can filter the packages based on that parameter,
    and the response will contain only the packages that match the filter. If the parameter value contains a comma,
    the filter will be an "in" query, meaning that the packages will be returned if the specified field contains any of the values in the parameter.
    ---
    tags:
      - Packages
    parameters:
      - name: package
        in: body
        required: false
        type: object
        description: Object containing the package filter
        schema:
          type: object
          properties:
            name:
              type: string
              description: The name of the package
            description:
              type: string
              description: The description of the package
            file_name:
              type: string
              description: The file name of the package, set by the user
            storage_file_name:
              type: string
              description: The name of the stored file, generated by the system
      - name: X-Limit
        in: header
        required: false
        type: integer
        description: The maximum number of packages to retrieve
      - name: X-Skip
        in: header
        required: false
        type: integer
        description: The number of packages to skip
      - name: X-Sort-Key
        in: header
        required: false
        type: string
        description: The key to sort the packages by
      - name: X-Sort-Order
        in: header
        required: false
        type: string
        description: The order to sort the packages in ('asc' or 'desc')
    responses:
      200:
        description: A list of packages and the total count.
        schema:
            type: object
            properties:
                count:
                    type: integer
                result:
                    type: array
                    items:
                        $ref: '#/definitions/Package'
      500:
        description: An error occurred
        schema:
            type: object
            properties:
                error:
                    type: string
    """
    try:
        limit, skip, sort_key, sort_order, _ = get_typed_list_headers(request.headers)
        kwargs = get_post_data_to_kwargs(request)

        packages = package_store.list(limit, skip, sort_key, sort_order, use_typing=True, **kwargs)

        result = [package.__dict__ for package in packages["result"]]

        response = {"count": packages["count"], "result": result}

        return jsonify(response), 200
    except Exception:
        return jsonify({"message": "An unexpected error occurred"}), 500


@bp.route("/count", methods=["GET"])
@jwt_auth_required(role="admin")
def get_packages_count():
    """Package count
    Retrieves the count of packages based on the provided parameters.
    By specifying a parameter in the url, you can filter the packages based on that parameter,
    and the response will contain only the packages that match the filter.
    ---
    tags:
      - Packages
    parameters:
      - in: query
        name: name
        schema:
        type: string
        description: The name of the package
      - in: query
        name: description
        schema:
        type: string
        description: The description of the package
      - in: query
        name: file_name
        schema:
        type: string
        description: Display name of the file, set by the user
      - in: query
        name: storage_file_name
        schema:
        type: string
        description: Name of the stored file, generated by the system
    responses:
      200:
        description: A list of packages and the total count.
        schema:
            type: object
            properties:
                count:
                    type: integer
                result:
                    type: array
                    items:
                        $ref: '#/definitions/Package'
      500:
        description: An error occurred
        schema:
            type: object
            properties:
                error:
                    type: string

    """
    try:
        kwargs = request.args.to_dict()
        count = package_store.count(**kwargs)
        response = count
        return jsonify(response), 200
    except Exception:
        return jsonify({"message": "An unexpected error occurred"}), 500


@bp.route("/count", methods=["POST"])
@jwt_auth_required(role="admin")
def packages_count():
    """Package count
    Retrieves the count of packages based on the provided parameters.
    Much like the GET /packages/count endpoint, but allows for more complex queries.
    By specifying a parameter in the body, you can filter the packages based on that parameter,
    and the response will contain only the count of packages that match the filter. If the parameter value contains a comma,
    the filter will be an "in" query, meaning that the packages will be returned if the specified field contains any of the values in the parameter.
    ---
    tags:
      - Packages
    parameters:
      - name: package
        in: body
        required: false
        type: object
        description: Object containing the package filter
        schema:
          type: object
          properties:
            name:
              type: string
              description: The name of the package
            description:
              type: string
              description: The description of the package
            file_name:
              type: string
              description: The file name of the package, set by the user
            storage_file_name:
              type: string
              description: The name of the stored file, generated by the system
    responses:
      200:
        description: A list of packages and the total count.
        schema:
            type: object
            properties:
                count:
                    type: integer
                result:
                    type: array
                    items:
                        $ref: '#/definitions/Package'
      500:
        description: An error occurred
        schema:
            type: object
            properties:
                error:
                    type: string
    """
    try:
        kwargs = get_post_data_to_kwargs(request)
        count = package_store.count(**kwargs)
        response = count
        return jsonify(response), 200
    except Exception:
        return jsonify({"message": "An unexpected error occurred"}), 500


@bp.route("/<string:id>", methods=["GET"])
@jwt_auth_required(role="admin")
def get_package(id: str):
    """Get package
    Retrieves a package based on the provided id.
    ---
    tags:
        - Packages
    parameters:
      - name: id
        in: path
        required: true
        type: string
        description: The id or package property of the package
    responses:
        200:
            description: The package
            schema:
                $ref: '#/definitions/package'
        404:
            description: The package was not found
            schema:
                type: object
                properties:
                    message:
                        type: string
        500:
            description: An error occurred
            schema:
                type: object
                properties:
                    message:
                        type: string
    """
    try:
        use_typing: bool = get_use_typing(request.headers)
        package = package_store.get(id, use_typing=use_typing)

        response = package.__dict__ if use_typing else package

        return jsonify(response), 200
    except EntityNotFound:
        return jsonify({"message": f"Entity with id: {id} not found"}), 404
    except Exception:
        return jsonify({"message": "An unexpected error occurred"}), 500


@bp.route("/active", methods=["GET"])
@jwt_auth_required(role="admin")
def get_active_package():
    """Get active package
    Retrieves the active package
    ---
    tags:
        - Packages
    responses:
        200:
            description: The package
            schema:
                $ref: '#/definitions/package'
        404:
            description: The package was not found
            schema:
                type: object
                properties:
                    message:
                        type: string
        500:
            description: An error occurred
            schema:
                type: object
                properties:
                    message:
                        type: string
    """
    try:
        use_typing: bool = get_use_typing(request.headers)
        package = package_store.get_active(use_typing=use_typing)
        response = package.__dict__ if use_typing else package

        return jsonify(response), 200
    except EntityNotFound:
        return jsonify({"message": f"Entity with id: {id} not found"}), 404
    except Exception:
        return jsonify({"message": "An unexpected error occurred"}), 500



from typing import Any, Dict, List, Tuple

import pymongo
from pymongo.database import Database

from fedn.network.storage.statestore.stores.store import Store


class Round:
    def __init__(self, id: str, round_id: str, status: str, round_config: dict, combiners: List[dict], round_data: dict):
        self.id = id
        self.round_id = round_id
        self.status = status
        self.round_config = round_config
        self.combiners = combiners
        self.round_data = round_data

    def from_dict(data: dict) -> "Round":
        return Round(
            id=str(data["_id"]),
            round_id=data["round_id"] if "round_id" in data else None,
            status=data["status"] if "status" in data else None,
            round_config=data["round_config"] if "round_config" in data else None,
            combiners=data["combiners"] if "combiners" in data else None,
            round_data=data["round_data"] if "round_data" in data else None
        )


class RoundStore(Store[Round]):
    def __init__(self, database: Database, collection: str):
        super().__init__(database, collection)

    def get(self, id: str, use_typing: bool = False) -> Round:
        """Get an entity by id
        param id: The id of the entity
            type: str
        param use_typing: Whether to return the entity as a typed object or as a dict
            type: bool
        return: The entity
        """
        response = super().get(id, use_typing=use_typing)
        return Round.from_dict(response) if use_typing else response

    def update(self, id: str, item: Round) -> bool:
        raise NotImplementedError("Update not implemented for RoundStore")

    def add(self, item: Round)-> Tuple[bool, Any]:
        raise NotImplementedError("Add not implemented for RoundStore")

    def delete(self, id: str) -> bool:
        raise NotImplementedError("Delete not implemented for RoundStore")

    def list(self, limit: int, skip: int, sort_key: str, sort_order=pymongo.DESCENDING, use_typing: bool = False, **kwargs) -> Dict[int, List[Round]]:
        """List entities
        param limit: The maximum number of entities to return
            type: int
            description: The maximum number of entities to return
        param skip: The number of entities to skip
            type: int
            description: The number of entities to skip
        param sort_key: The key to sort by
            type: str
            description: The key to sort by
        param sort_order: The order to sort by
            type: pymongo.DESCENDING
            description: The order to sort by
        param use_typing: Whether to return the entity as a typed object or as a dict
            type: bool
        return: The entities
        """
        response = super().list(limit, skip, sort_key or "round_id", sort_order, use_typing=use_typing, **kwargs)

        result = [Round.from_dict(item) for item in response["result"]] if use_typing else response["result"]

        return {
            "count": response["count"],
            "result": result
        }



from fedn.network.api.v1.client_routes import bp as client_bp
from fedn.network.api.v1.combiner_routes import bp as combiner_bp
from fedn.network.api.v1.inference_routes import bp as inference_bp
from fedn.network.api.v1.model_routes import bp as model_bp
from fedn.network.api.v1.package_routes import bp as package_bp
from fedn.network.api.v1.round_routes import bp as round_bp
from fedn.network.api.v1.session_routes import bp as session_bp
from fedn.network.api.v1.status_routes import bp as status_bp
from fedn.network.api.v1.validation_routes import bp as validation_bp

_routes = [client_bp, combiner_bp, model_bp, package_bp, round_bp, session_bp, status_bp, validation_bp, inference_bp]



import os
import unittest
from unittest.mock import MagicMock, patch

from fedn.common.storage.models.tempmodelstorage import TempModelStorage


class TestTempModelStorage(unittest.TestCase):

    def setUp(self):
        # Setup mock for os.environ.get for FEDN_MODEL_DIR
        self.patcher = patch('os.environ.get')
        self.mock_get = self.patcher.start()
        # Return value of mock should same folder as this file
        self.mock_get.return_value = os.path.dirname(os.path.realpath(__file__))

        # Setup storage
        self.storage = TempModelStorage()

        # add mock data to storage dicts
        self.storage.models = {"model_id1": "model1", "model_id2": "model2"}
        self.storage.models_metadata = {"model_id1": "model1", "model_id2": "model2"}

        # Create mock file as BytesIO object
        self.mock_file = MagicMock()
        self.mock_file.read.return_value = "model1"
        self.mock_file.seek.return_value = 0
        self.mock_file.write.return_value = None

    # Test that the storage is initialized with the correct default directory and data structures
    def test_init(self):
        self.assertEqual(self.storage.default_dir, os.path.dirname(os.path.realpath(__file__)))
        self.assertEqual(self.storage.models, {"model_id1": "model1", "model_id2": "model2"})
        self.assertEqual(self.storage.models_metadata, {"model_id1": "model1", "model_id2": "model2"})

    # Test that the storage can get a model

    def test_get(self):
        """ Test that the storage can get a model """

        # Test that it returns None if model_id does not exist
        self.assertEqual(self.storage.get("model_id3"), None)

        # TODO: Patch fedn.ModelStatus.OK and open to return True and mock_file respectively

    def test_get_metadata(self):
        """ Test that the storage can get a model metadata """

        # Test that it returns KeyError if model_id does not exist
        with self.assertRaises(KeyError):
            self.storage.get_meta("model_id3")

        # Test that it returns the correct metadata if model_id exists
        self.assertEqual(self.storage.get_meta("model_id1"), "model1")

    def test_set_meta(self):
        """ Test that the storage can set a model metadata """

        # Test that it returns the correct metadata if model_id exists
        self.storage.set_meta("model_id1", "model3")
        self.assertEqual(self.storage.get_meta("model_id1"), "model3")

    def test_delete(self):
        """ Test that the storage can delete a model """

        # Test that it returns False if model_id does not exist
        self.assertEqual(self.storage.delete("model_id3"), False)

        # Patch os.remove to return True
        with patch('os.remove', return_value=True) as mock_remove:

            # Test that it returns True if model_id exists
            self.assertEqual(self.storage.delete("model_id1"), True)

            # Test that os.remove is called with the correct path
            mock_remove.assert_called_with(os.path.join(self.storage.default_dir, "model_id1"))

            # Test that the model is removed from the storage
            self.assertEqual(self.storage.models, {"model_id2": "model2"})

            # Test that the model metadata is removed from the storage
            self.assertEqual(self.storage.models_metadata, {"model_id2": "model2"})

    def test_delete_all(self):
        """ Test that the storage can delete all models """

        # Patch os.remove to return True
        with patch('os.remove', return_value=True) as mock_remove:

            # Test that it returns True if model_id exists
            self.assertEqual(self.storage.delete_all(), True)

            # Test that os.remove is called with the correct path
            mock_remove.assert_called_with(os.path.join(self.storage.default_dir, "model_id2"))

            # Test that the model is removed from the storage
            self.assertEqual(self.storage.models, {})

            # Test that the model metadata is removed from the storage
            self.assertEqual(self.storage.models_metadata, {})


if __name__ == '__main__':
    unittest.main()



import threading

from flask import Blueprint, jsonify, request

from fedn.network.api.auth import jwt_auth_required
from fedn.network.api.shared import control
from fedn.network.api.v1.shared import api_version

bp = Blueprint("inference", __name__, url_prefix=f"/api/{api_version}/infer")


@bp.route("/start", methods=["POST"])
@jwt_auth_required(role="admin")
def start_session():
    """Start a new inference session.
    param: session_id: The session id to start.
    type: session_id: str
    param: rounds: The number of rounds to run.
    type: rounds: int
    """
    try:
        data = request.json if request.headers["Content-Type"] == "application/json" else request.form.to_dict()
        session_id: str = data.get("session_id")

        if not session_id or session_id == "":
            return jsonify({"message": "Session ID is required"}), 400

        session_config = {"session_id": session_id}

        threading.Thread(target=control.inference_session, kwargs={"config": session_config}).start()

        return jsonify({"message": "Inference session started"}), 200
    except Exception:
        return jsonify({"message": "Failed to start inference session"}), 500



from typing import Tuple

from flwr.client import ClientApp
from flwr.common import (
    Context,
    EvaluateIns,
    FitIns,
    GetParametersIns,
    Message,
    MessageType,
    MessageTypeLegacy,
    Metadata,
    NDArrays,
    ndarrays_to_parameters,
    parameters_to_ndarrays,
)
from flwr.common.recordset_compat import (
    evaluateins_to_recordset,
    fitins_to_recordset,
    getparametersins_to_recordset,
    recordset_to_evaluateres,
    recordset_to_fitres,
    recordset_to_getparametersres,
)


class FlwrClientAppAdapter:
    """Flwr ClientApp wrapper."""

    def __init__(self, app: ClientApp) -> None:
        self.app = app

    def init_parameters(self, partition_id: int, config: dict = {}):
        # Construct a get_parameters message for the ClientApp
        message, context = self._construct_message(MessageTypeLegacy.GET_PARAMETERS, [], partition_id, config)
        # Call client app with train message
        client_return_message = self.app(message, context)
        # return NDArrays of clients parameters
        parameters = self._parse_get_parameters_message(client_return_message)
        if len(parameters) == 0:
            raise ValueError(
                "The 'parameters' list is empty. Ensure your flower \
                             client has implemented a get_parameters() function."
            )
        return parameters

    def train(self, parameters: NDArrays, partition_id: int, config: dict = {}):
        # Construct a train message for the ClientApp with given parameters
        message, context = self._construct_message(MessageType.TRAIN, parameters, partition_id, config)
        # Call client app with train message
        client_return_message = self.app(message, context)
        # Parse return message
        params, num_examples = self._parse_train_message(client_return_message)
        return params, num_examples

    def evaluate(self, parameters: NDArrays, partition_id: int, config: dict = {}):
        # Construct an evaluate message for the ClientApp with given parameters
        message, context = self._construct_message(MessageType.EVALUATE, parameters, partition_id, config)
        # Call client app with evaluate message
        client_return_message = self.app(message, context)
        # Parse return message
        loss, accuracy = self._parse_evaluate_message(client_return_message)
        return loss, accuracy

    def _parse_get_parameters_message(self, message: Message) -> NDArrays:
        get_parameters_res = recordset_to_getparametersres(message.content, keep_input=False)
        return parameters_to_ndarrays(get_parameters_res.parameters)

    def _parse_train_message(self, message: Message) -> Tuple[NDArrays, int]:
        fitres = recordset_to_fitres(message.content, keep_input=False)
        params, num_examples = (
            parameters_to_ndarrays(fitres.parameters),
            fitres.num_examples,
        )
        return params, num_examples

    def _parse_evaluate_message(self, message: Message) -> Tuple[float, float]:
        evaluateres = recordset_to_evaluateres(message.content)
        return evaluateres.loss, evaluateres.metrics.get("accuracy", -1)

    def _construct_message(
        self,
        message_type: MessageType,
        parameters: NDArrays,
        partition_id: int,
        config: dict,
    ) -> Tuple[Message, Context]:
        parameters = ndarrays_to_parameters(parameters)
        if message_type == MessageType.TRAIN:
            fit_ins: FitIns = FitIns(parameters=parameters, config=config)
            recordset = fitins_to_recordset(fitins=fit_ins, keep_input=False)
        if message_type == MessageType.EVALUATE:
            ev_ins: EvaluateIns = EvaluateIns(parameters=parameters, config=config)
            recordset = evaluateins_to_recordset(evaluateins=ev_ins, keep_input=False)
        if message_type == MessageTypeLegacy.GET_PARAMETERS:
            get_parameters_ins: GetParametersIns = GetParametersIns(config=config)
            recordset = getparametersins_to_recordset(getparameters_ins=get_parameters_ins)

        metadata = Metadata(
            run_id=0,
            message_id="",
            src_node_id=0,
            dst_node_id=0,
            reply_to_message="",
            group_id="",
            ttl=0.0,
            message_type=message_type,
            partition_id=partition_id,
        )
        context = Context(recordset)
        message = Message(metadata=metadata, content=recordset)
        return message, context



# This file contains the PackageRuntime class, which is used to download, validate and unpack compute packages.
#
#
import cgi
import os
import tarfile

import requests

from fedn.common.config import FEDN_AUTH_SCHEME, FEDN_CUSTOM_URL_PREFIX
from fedn.common.log_config import logger
from fedn.utils.checksum import sha
from fedn.utils.dispatcher import Dispatcher, _read_yaml_file


class PackageRuntime:
    """PackageRuntime is used to download, validate and unpack compute packages.

    :param package_path: path to compute package
    :type package_path: str
    :param package_dir: directory to unpack compute package
    :type package_dir: str
    """

    def __init__(self, package_path):
        self.dispatch_config = {
            "entry_points": {
                "predict": {"command": "python3 predict.py"},
                "train": {"command": "python3 train.py"},
                "validate": {"command": "python3 validate.py"},
            }
        }

        self.pkg_path = package_path
        self.pkg_name = None
        self.checksum = None
        self.expected_checksum = None

    def download(self, host, port, token, force_ssl=False, secure=False, name=None):
        """Download compute package from controller

        :param host: host of controller
        :param port: port of controller
        :param token: token for authentication
        :param name: name of package
        :return: True if download was successful, None otherwise
        :rtype: bool
        """
        # for https we assume a an ingress handles permanent redirect (308)
        if force_ssl:
            scheme = "https"
        else:
            scheme = "http"
        if port:
            path = f"{scheme}://{host}:{port}{FEDN_CUSTOM_URL_PREFIX}/download_package"
        else:
            path = f"{scheme}://{host}{FEDN_CUSTOM_URL_PREFIX}/download_package"
        if name:
            path = path + "?name={}".format(name)

        with requests.get(path, stream=True, verify=False, headers={"Authorization": f"{FEDN_AUTH_SCHEME} {token}"}) as r:
            if 200 <= r.status_code < 204:
                params = cgi.parse_header(r.headers.get("Content-Disposition", ""))[-1]
                try:
                    self.pkg_name = params["filename"]
                except KeyError:
                    logger.error("No package returned.")
                    return None
                r.raise_for_status()
                with open(os.path.join(self.pkg_path, self.pkg_name), "wb") as f:
                    for chunk in r.iter_content(chunk_size=8192):
                        f.write(chunk)
        if port:
            path = f"{scheme}://{host}:{port}{FEDN_CUSTOM_URL_PREFIX}/get_package_checksum"
        else:
            path = f"{scheme}://{host}{FEDN_CUSTOM_URL_PREFIX}/get_package_checksum"

        if name:
            path = path + "?name={}".format(name)
        with requests.get(path, verify=False, headers={"Authorization": f"{FEDN_AUTH_SCHEME} {token}"}) as r:
            if 200 <= r.status_code < 204:
                data = r.json()
                try:
                    self.checksum = data["checksum"]
                except Exception:
                    logger.error("Could not extract checksum.")

        return True

    def validate(self, expected_checksum):
        """Validate the package against the checksum provided by the controller

        :param expected_checksum: checksum provided by the controller
        :return: True if checksums match, False otherwise
        :rtype: bool
        """
        self.expected_checksum = expected_checksum

        # crosscheck checksum and unpack if security checks are ok.
        file_checksum = str(sha(os.path.join(self.pkg_path, self.pkg_name)))

        if self.checksum == self.expected_checksum == file_checksum:
            logger.info("Package validated {}".format(self.checksum))
            return True
        else:
            return False

    def unpack(self):
        """Unpack the compute package

        :return: True if unpacking was successful, False otherwise
        :rtype: bool
        """
        if self.pkg_name:
            f = None
            if self.pkg_name.endswith("tar.gz"):
                f = tarfile.open(os.path.join(self.pkg_path, self.pkg_name), "r:gz")
            if self.pkg_name.endswith(".tgz"):
                f = tarfile.open(os.path.join(self.pkg_path, self.pkg_name), "r:gz")
            if self.pkg_name.endswith("tar.bz2"):
                f = tarfile.open(os.path.join(self.pkg_path, self.pkg_name), "r:bz2")
        else:
            logger.error("Failed to unpack compute package, no pkg_name set." "Has the reducer been configured with a compute package?")
            return False

        try:
            if f:
                f.extractall(self.pkg_path)
                logger.info("Successfully extracted compute package content in {}".format(self.pkg_path))
                # delete the tarball
                logger.info("Deleting temporary package tarball file.")
                f.close()
                os.remove(os.path.join(self.pkg_path, self.pkg_name))
                # search for file fedn.yaml in extracted package
                for root, dirs, files in os.walk(self.pkg_path):
                    if "fedn.yaml" in files:
                        # Get the path to where fedn.yaml is located
                        logger.info("Found fedn.yaml file in {}".format(root))
                        return True, root

                logger.error("No fedn.yaml file found in extracted package!")
                return False, ""
        except Exception:
            logger.error("Error extracting files.")
            # delete the tarball
            os.remove(os.path.join(self.pkg_path, self.pkg_name))
            return False, ""

    def dispatcher(self, run_path):
        """Dispatch the compute package

        :param run_path: path to dispatch the compute package
        :type run_path: str
        :return: Dispatcher object
        :rtype: :class:`fedn.utils.dispatcher.Dispatcher`
        """
        self.dispatch_config = _read_yaml_file(os.path.join(run_path, "fedn.yaml"))
        dispatcher = Dispatcher(self.dispatch_config, run_path)

        return dispatcher



import os

from flask import Flask, jsonify, request

from fedn.common.config import get_controller_config
from fedn.network.api.auth import jwt_auth_required
from fedn.network.api.interface import API
from fedn.network.api.shared import control, statestore
from fedn.network.api.v1 import _routes

custom_url_prefix = os.environ.get("FEDN_CUSTOM_URL_PREFIX", False)
# statestore_config,modelstorage_config,network_id,control=set_statestore_config()
api = API(statestore, control)
app = Flask(__name__)
for bp in _routes:
    app.register_blueprint(bp)
    if custom_url_prefix:
        app.register_blueprint(bp, name=f"{bp.name}_custom", url_prefix=f"{custom_url_prefix}{bp.url_prefix}")


@app.route("/health", methods=["GET"])
def health_check():
    return "OK", 200


if custom_url_prefix:
    app.add_url_rule(f"{custom_url_prefix}/health", view_func=health_check, methods=["GET"])


@app.route("/get_model_trail", methods=["GET"])
@jwt_auth_required(role="admin")
def get_model_trail():
    """Get the model trail for a given session.
    param: session: The session id to get the model trail for.
    type: session: str
    return: The model trail for the given session as a json object.
    rtype: json
    """
    return api.get_model_trail()


if custom_url_prefix:
    app.add_url_rule(f"{custom_url_prefix}/get_model_trail", view_func=get_model_trail, methods=["GET"])


@app.route("/get_model_ancestors", methods=["GET"])
@jwt_auth_required(role="admin")
def get_model_ancestors():
    """Get the ancestors of a model.
    param: model: The model id to get the ancestors for.
    type: model: str
    param: limit: The maximum number of ancestors to return.
    type: limit: int
    return: A list of model objects that the model derives from.
    rtype: json
    """
    model = request.args.get("model", None)
    limit = request.args.get("limit", None)

    return api.get_model_ancestors(model, limit)


if custom_url_prefix:
    app.add_url_rule(f"{custom_url_prefix}/get_model_ancestors", view_func=get_model_ancestors, methods=["GET"])


@app.route("/get_model_descendants", methods=["GET"])
@jwt_auth_required(role="admin")
def get_model_descendants():
    """Get the ancestors of a model.
    param: model: The model id to get the child for.
    type: model: str
    param: limit: The maximum number of descendants to return.
    type: limit: int
    return: A list of model objects that are descendents of the provided model id.
    rtype: json
    """
    model = request.args.get("model", None)
    limit = request.args.get("limit", None)

    return api.get_model_descendants(model, limit)


if custom_url_prefix:
    app.add_url_rule(f"{custom_url_prefix}/get_model_descendants", view_func=get_model_descendants, methods=["GET"])


@app.route("/list_models", methods=["GET"])
@jwt_auth_required(role="admin")
def list_models():
    """Get models from the statestore.
    param:
    session_id: The session id to get the model trail for.
    limit: The maximum number of models to return.
    type: limit: int
    param: skip: The number of models to skip.
    type: skip: int
    Returns:
        _type_: json
    """
    session_id = request.args.get("session_id", None)
    limit = request.args.get("limit", None)
    skip = request.args.get("skip", None)
    include_active = request.args.get("include_active", None)

    return api.get_models(session_id, limit, skip, include_active)


if custom_url_prefix:
    app.add_url_rule(f"{custom_url_prefix}/list_models", view_func=list_models, methods=["GET"])


@app.route("/get_model", methods=["GET"])
@jwt_auth_required(role="admin")
def get_model():
    """Get a model from the statestore.
    param: model: The model id to get.
    type: model: str
    return: The model as a json object.
    rtype: json
    """
    model = request.args.get("model", None)
    if model is None:
        return jsonify({"success": False, "message": "Missing model id."}), 400

    return api.get_model(model)


if custom_url_prefix:
    app.add_url_rule(f"{custom_url_prefix}/get_model", view_func=get_model, methods=["GET"])


@app.route("/delete_model_trail", methods=["GET", "POST"])
@jwt_auth_required(role="admin")
def delete_model_trail():
    """Delete the model trail for a given session.
    param: session: The session id to delete the model trail for.
    type: session: str
    return: The response from the statestore.
    rtype: json
    """
    return jsonify({"message": "Not implemented"}), 501


if custom_url_prefix:
    app.add_url_rule(f"{custom_url_prefix}/delete_model_trail", view_func=delete_model_trail, methods=["GET", "POST"])


@app.route("/list_clients", methods=["GET"])
@jwt_auth_required(role="admin")
def list_clients():
    """Get all clients from the statestore.
    return: All clients as a json object.
    rtype: json
    """
    limit = request.args.get("limit", None)
    skip = request.args.get("skip", None)
    status = request.args.get("status", None)

    return api.get_clients(limit, skip, status)


if custom_url_prefix:
    app.add_url_rule(f"{custom_url_prefix}/list_clients", view_func=list_clients, methods=["GET"])


@app.route("/get_active_clients", methods=["GET"])
@jwt_auth_required(role="admin")
def get_active_clients():
    """Get all active clients from the statestore.
    param: combiner_id: The combiner id to get active clients for.
    type: combiner_id: str
    return: All active clients as a json object.
    rtype: json
    """
    combiner_id = request.args.get("combiner", None)
    if combiner_id is None:
        return (
            jsonify({"success": False, "message": "Missing combiner id."}),
            400,
        )
    return api.get_active_clients(combiner_id)


if custom_url_prefix:
    app.add_url_rule(f"{custom_url_prefix}/get_active_clients", view_func=get_active_clients, methods=["GET"])


@app.route("/list_combiners", methods=["GET"])
@jwt_auth_required(role="admin")
def list_combiners():
    """Get all combiners in the network.
    return: All combiners as a json object.
    rtype: json
    """
    limit = request.args.get("limit", None)
    skip = request.args.get("skip", None)

    return api.get_all_combiners(limit, skip)


if custom_url_prefix:
    app.add_url_rule(f"{custom_url_prefix}/list_combiners", view_func=list_combiners, methods=["GET"])


@app.route("/get_combiner", methods=["GET"])
@jwt_auth_required(role="admin")
def get_combiner():
    """Get a combiner from the statestore.
    param: combiner_id: The combiner id to get.
    type: combiner_id: str
    return: The combiner as a json object.
    rtype: json
    """
    combiner_id = request.args.get("combiner", None)
    if combiner_id is None:
        return (
            jsonify({"success": False, "message": "Missing combiner id."}),
            400,
        )
    return api.get_combiner(combiner_id)


if custom_url_prefix:
    app.add_url_rule(f"{custom_url_prefix}/get_combiner", view_func=get_combiner, methods=["GET"])


@app.route("/list_rounds", methods=["GET"])
@jwt_auth_required(role="admin")
def list_rounds():
    """Get all rounds from the statestore.
    return: All rounds as a json object.
    rtype: json
    """
    return api.get_all_rounds()


if custom_url_prefix:
    app.add_url_rule(f"{custom_url_prefix}/list_rounds", view_func=list_rounds, methods=["GET"])


@app.route("/get_round", methods=["GET"])
@jwt_auth_required(role="admin")
def get_round():
    """Get a round from the statestore.
    param: round_id: The round id to get.
    type: round_id: str
    return: The round as a json object.
    rtype: json
    """
    round_id = request.args.get("round_id", None)
    if round_id is None:
        return jsonify({"success": False, "message": "Missing round id."}), 400
    return api.get_round(round_id)


if custom_url_prefix:
    app.add_url_rule(f"{custom_url_prefix}/get_round", view_func=get_round, methods=["GET"])


@app.route("/start_session", methods=["GET", "POST"])
@jwt_auth_required(role="admin")
def start_session():
    """Start a new session.
    return: The response from control.
    rtype: json
    """
    json_data = request.get_json()
    return api.start_session(**json_data)


if custom_url_prefix:
    app.add_url_rule(f"{custom_url_prefix}/start_session", view_func=start_session, methods=["GET", "POST"])


@app.route("/list_sessions", methods=["GET"])
@jwt_auth_required(role="admin")
def list_sessions():
    """Get all sessions from the statestore.
    return: All sessions as a json object.
    rtype: json
    """
    limit = request.args.get("limit", None)
    skip = request.args.get("skip", None)

    return api.get_all_sessions(limit, skip)


if custom_url_prefix:
    app.add_url_rule(f"{custom_url_prefix}/list_sessions", view_func=list_sessions, methods=["GET"])


@app.route("/get_session", methods=["GET"])
@jwt_auth_required(role="admin")
def get_session():
    """Get a session from the statestore.
    param: session_id: The session id to get.
    type: session_id: str
    return: The session as a json object.
    rtype: json
    """
    session_id = request.args.get("session_id", None)
    if session_id is None:
        return (
            jsonify({"success": False, "message": "Missing session id."}),
            400,
        )
    return api.get_session(session_id)


if custom_url_prefix:
    app.add_url_rule(f"{custom_url_prefix}/get_session", view_func=get_session, methods=["GET"])


@app.route("/set_active_package", methods=["PUT"])
@jwt_auth_required(role="admin")
def set_active_package():
    id = request.args.get("id", None)
    return api.set_active_compute_package(id)


if custom_url_prefix:
    app.add_url_rule(f"{custom_url_prefix}/set_active_package", view_func=set_active_package, methods=["PUT"])


@app.route("/set_package", methods=["POST"])
@jwt_auth_required(role="admin")
def set_package():
    """ Set the compute package in the statestore.
        Usage with curl:
        curl -k -X POST \
            -F file=@package.tgz \
            -F helper="kerashelper" \
            http://localhost:8092/set_package

    param: file: The compute package file to set.
    type: file: file
    return: The response from the statestore.
    rtype: json
    """
    helper_type = request.form.get("helper", None)
    name = request.form.get("name", None)
    description = request.form.get("description", None)

    if helper_type is None:
        return (
            jsonify({"success": False, "message": "Missing helper type."}),
            400,
        )
    try:
        file = request.files["file"]
    except KeyError:
        return jsonify({"success": False, "message": "Missing file."}), 400
    return api.set_compute_package(file=file, helper_type=helper_type, name=name, description=description)


if custom_url_prefix:
    app.add_url_rule(f"{custom_url_prefix}/set_package", view_func=set_package, methods=["POST"])


@app.route("/get_package", methods=["GET"])
@jwt_auth_required(role="admin")
def get_package():
    """Get the compute package from the statestore.
    return: The compute package as a json object.
    rtype: json
    """
    return api.get_compute_package()


if custom_url_prefix:
    app.add_url_rule(f"{custom_url_prefix}/get_package", view_func=get_package, methods=["GET"])


@app.route("/list_compute_packages", methods=["GET"])
@jwt_auth_required(role="admin")
def list_compute_packages():
    """Get the compute package from the statestore.
    return: The compute package as a json object.
    rtype: json
    """
    limit = request.args.get("limit", None)
    skip = request.args.get("skip", None)
    include_active = request.args.get("include_active", None)

    return api.list_compute_packages(limit=limit, skip=skip, include_active=include_active)


if custom_url_prefix:
    app.add_url_rule(f"{custom_url_prefix}/list_compute_packages", view_func=list_compute_packages, methods=["GET"])


@app.route("/download_package", methods=["GET"])
@jwt_auth_required(role="client")
def download_package():
    """Download the compute package.
    return: The compute package as a json object.
    rtype: json
    """
    name = request.args.get("name", None)
    return api.download_compute_package(name)


if custom_url_prefix:
    app.add_url_rule(f"{custom_url_prefix}/download_package", view_func=download_package, methods=["GET"])


@app.route("/get_package_checksum", methods=["GET"])
@jwt_auth_required(role="client")
def get_package_checksum():
    name = request.args.get("name", None)
    return api.get_checksum(name)


if custom_url_prefix:
    app.add_url_rule(f"{custom_url_prefix}/get_package_checksum", view_func=get_package_checksum, methods=["GET"])


@app.route("/get_latest_model", methods=["GET"])
@jwt_auth_required(role="admin")
def get_latest_model():
    """Get the latest model from the statestore.
    return: The initial model as a json object.
    rtype: json
    """
    return api.get_latest_model()


if custom_url_prefix:
    app.add_url_rule(f"{custom_url_prefix}/get_latest_model", view_func=get_latest_model, methods=["GET"])


@app.route("/set_current_model", methods=["PUT"])
@jwt_auth_required(role="admin")
def set_current_model():
    """Set the initial model in the statestore and upload to model repository.
        Usage with curl:
        curl -k -X PUT
            -F id=<model-id>
            http://localhost:8092/set_current_model

    param: id: The model id to set.
    type: id: str
    return: boolean.
    rtype: json
    """
    id = request.args.get("id", None)
    if id is None:
        return jsonify({"success": False, "message": "Missing model id."}), 400
    return api.set_current_model(id)


if custom_url_prefix:
    app.add_url_rule(f"{custom_url_prefix}/set_current_model", view_func=set_current_model, methods=["PUT"])

# Get initial model endpoint


@app.route("/get_initial_model", methods=["GET"])
@jwt_auth_required(role="admin")
def get_initial_model():
    """Get the initial model from the statestore.
    return: The initial model as a json object.
    rtype: json
    """
    return api.get_initial_model()


if custom_url_prefix:
    app.add_url_rule(f"{custom_url_prefix}/get_initial_model", view_func=get_initial_model, methods=["GET"])


@app.route("/set_initial_model", methods=["POST"])
@jwt_auth_required(role="admin")
def set_initial_model():
    """Set the initial model in the statestore and upload to model repository.
        Usage with curl:
        curl -k -X POST
            -F file=@seed.npz
            http://localhost:8092/set_initial_model

    param: file: The initial model file to set.
    type: file: file
    return: The response from the statestore.
    rtype: json
    """
    try:
        file = request.files["file"]
    except KeyError:
        return jsonify({"success": False, "message": "Missing file."}), 400
    return api.set_initial_model(file)


if custom_url_prefix:
    app.add_url_rule(f"{custom_url_prefix}/set_initial_model", view_func=set_initial_model, methods=["POST"])


@app.route("/get_controller_status", methods=["GET"])
@jwt_auth_required(role="admin")
def get_controller_status():
    """Get the status of the controller.
    return: The status as a json object.
    rtype: json
    """
    return api.get_controller_status()


if custom_url_prefix:
    app.add_url_rule(f"{custom_url_prefix}/get_controller_status", view_func=get_controller_status, methods=["GET"])


@app.route("/get_client_config", methods=["GET"])
@jwt_auth_required(role="admin")
def get_client_config():
    """Get the client configuration.
    return: The client configuration as a json object.
    rtype: json
    """
    checksum_arg = request.args.get("checksum", "true")
    checksum = checksum_arg.lower() != "false"
    return api.get_client_config(checksum)


if custom_url_prefix:
    app.add_url_rule(f"{custom_url_prefix}/get_client_config", view_func=get_client_config, methods=["GET"])


@app.route("/get_events", methods=["GET"])
@jwt_auth_required(role="admin")
def get_events():
    """Get the events from the statestore.
    return: The events as a json object.
    rtype: json
    """
    # TODO: except filter with request.get_json()
    kwargs = request.args.to_dict()

    return api.get_events(**kwargs)


if custom_url_prefix:
    app.add_url_rule(f"{custom_url_prefix}/get_events", view_func=get_events, methods=["GET"])


@app.route("/list_validations", methods=["GET"])
@jwt_auth_required(role="admin")
def list_validations():
    """Get all validations from the statestore.
    return: All validations as a json object.
    rtype: json
    """
    # TODO: except filter with request.get_json()
    kwargs = request.args.to_dict()
    return api.get_all_validations(**kwargs)


if custom_url_prefix:
    app.add_url_rule(f"{custom_url_prefix}/list_validations", view_func=list_validations, methods=["GET"])


@app.route("/add_combiner", methods=["POST"])
@jwt_auth_required(role="combiner")
def add_combiner():
    """Add a combiner to the network.
    return: The response from the statestore.
    rtype: json
    """
    json_data = request.get_json()
    remote_addr = request.remote_addr
    try:
        response = api.add_combiner(**json_data, remote_addr=remote_addr)
    except TypeError:
        return jsonify({"success": False, "message": "Invalid data provided"}), 400
    except Exception:
        return jsonify({"success": False, "message": "An unexpected error occurred"}), 500
    return response


if custom_url_prefix:
    app.add_url_rule(f"{custom_url_prefix}/add_combiner", view_func=add_combiner, methods=["POST"])


@app.route("/add_client", methods=["POST"])
@jwt_auth_required(role="client")
def add_client():
    """Add a client to the network.
    return: The response from control.
    rtype: json
    """
    json_data = request.get_json()
    remote_addr = request.remote_addr
    try:
        response = api.add_client(**json_data, remote_addr=remote_addr)
    except TypeError:
        return jsonify({"success": False, "message": "Invalid data provided"}), 400
    except Exception:
        return jsonify({"success": False, "message": "An unexpected error occurred"}), 500
    return response


if custom_url_prefix:
    app.add_url_rule(f"{custom_url_prefix}/add_client", view_func=add_client, methods=["POST"])


@app.route("/list_combiners_data", methods=["POST"])
@jwt_auth_required(role="admin")
def list_combiners_data():
    """List data from combiners.
    return: The response from control.
    rtype: json
    """
    json_data = request.get_json()

    # expects a list of combiner names (strings) in an array
    combiners = json_data.get("combiners", None)

    try:
        response = api.list_combiners_data(combiners)
    except TypeError:
        return jsonify({"success": False, "message": "Invalid data provided"}), 400
    except Exception:
        return jsonify({"success": False, "message": "An unexpected error occurred"}), 500
    return response


if custom_url_prefix:
    app.add_url_rule(f"{custom_url_prefix}/list_combiners_data", view_func=list_combiners_data, methods=["POST"])


def start_server_api():
    config = get_controller_config()
    port = config["port"]
    debug = config["debug"]
    host = "0.0.0.0"
    app.run(debug=debug, port=port, host=host)


if __name__ == "__main__":
    start_server_api()



from abc import ABC, abstractmethod


class LoadBalancerBase(ABC):
    """Abstract base class for load balancers.

    :param network: A handle to the network.
    :type network: class: `fedn.network.api.network.Network`
    """

    def __init__(self, network):
        """ """
        self.network = network

    @abstractmethod
    def find_combiner(self):
        """Find a combiner to connect to."""
        pass



import threading

from flask import Blueprint, jsonify, request

from fedn.network.api.auth import jwt_auth_required
from fedn.network.api.shared import control
from fedn.network.api.v1.shared import api_version, get_post_data_to_kwargs, get_typed_list_headers, mdb
from fedn.network.combiner.interfaces import CombinerUnavailableError
from fedn.network.state import ReducerState
from fedn.network.storage.statestore.stores.session_store import SessionStore
from fedn.network.storage.statestore.stores.shared import EntityNotFound

from .model_routes import model_store

bp = Blueprint("session", __name__, url_prefix=f"/api/{api_version}/sessions")

session_store = SessionStore(mdb, "control.sessions")


@bp.route("/", methods=["GET"])
@jwt_auth_required(role="admin")
def get_sessions():
    """Get sessions
    Retrieves a list of sessions based on the provided parameters.
    By specifying a parameter in the url, you can filter the sessions based on that parameter,
    and the response will contain only the sessions that match the filter.
    ---
    tags:
        - Sessions
    parameters:
      - name: status
        in: query
        required: false
        type: string
        description: Status of the session
      - name: session_id
        in: query
        required: false
        type: string
      - name: X-Limit
        in: header
        required: false
        type: integer
        description: The maximum number of sessions to retrieve
      - name: X-Skip
        in: header
        required: false
        type: integer
        description: The number of sessions to skip
      - name: X-Sort-Key
        in: header
        required: false
        type: string
        description: The key to sort the sessions by
      - name: X-Sort-Order
        in: header
        required: false
        type: string
        description: The order to sort the sessions in ('asc' or 'desc')
    definitions:
      Session:
        type: object
        properties:
          id:
            type: string
          status:
            type: string
          session_id:
            type: string
          session_config:
            type: object
    responses:
      200:
        description: A list of sessions and the total count.
        schema:
            type: object
            properties:
                count:
                    type: integer
                result:
                    type: array
                    items:
                        $ref: '#/definitions/Session'
      500:
        description: An error occurred
        schema:
            type: object
            properties:
                message:
                    type: string
    """
    try:
        limit, skip, sort_key, sort_order, _ = get_typed_list_headers(request.headers)
        kwargs = request.args.to_dict()

        sessions = session_store.list(limit, skip, sort_key, sort_order, use_typing=False, **kwargs)

        result = sessions["result"]

        response = {"count": sessions["count"], "result": result}

        return jsonify(response), 200
    except Exception:
        return jsonify({"message": "An unexpected error occurred"}), 500


@bp.route("/list", methods=["POST"])
@jwt_auth_required(role="admin")
def list_sessions():
    """List sessions
    Retrieves a list of sessions based on the provided parameters.
    Works much like the GET /sessions endpoint, but allows for more complex queries.
    By specifying a parameter in the request body, you can filter the sessions based on that parameter,
    and the response will contain only the sessions that match the filter. If the parameter value contains a comma,
    the filter will be an "in" query, meaning that the sessions will be returned if the specified field contains any of the values in the parameter.
    ---
    tags:
      - Sessions
    parameters:
      - name: session
        in: body
        required: false
        schema:
          type: object
          properties:
            status:
              type: string
            session_id:
              type: string
      - name: X-Limit
        in: header
        required: false
        type: integer
        description: The maximum number of sessions to retrieve
      - name: X-Skip
        in: header
        required: false
        type: integer
        description: The number of sessions to skip
      - name: X-Sort-Key
        in: header
        required: false
        type: string
        description: The key to sort the sessions by
      - name: X-Sort-Order
        in: header
        required: false
        type: string
        description: The order to sort the sessions in ('asc' or 'desc')
    responses:
      200:
        description: A list of sessions and the total count.
        schema:
            type: object
            properties:
                count:
                    type: integer
                result:
                    type: array
                    items:
                        $ref: '#/definitions/Session'
      500:
        description: An error occurred
        schema:
            type: object
            properties:
                message:
                    type: string
    """
    try:
        limit, skip, sort_key, sort_order, _ = get_typed_list_headers(request.headers)
        kwargs = get_post_data_to_kwargs(request)

        sessions = session_store.list(limit, skip, sort_key, sort_order, use_typing=False, **kwargs)

        result = sessions["result"]

        response = {"count": sessions["count"], "result": result}

        return jsonify(response), 200
    except Exception:
        return jsonify({"message": "An unexpected error occurred"}), 500


@bp.route("/count", methods=["GET"])
@jwt_auth_required(role="admin")
def get_sessions_count():
    """Sessions count
    Retrieves the count of sessions based on the provided parameters.
    By specifying a parameter in the url, you can filter the sessions based on that parameter,
    and the response will contain only the count of sessions that match the filter.
    ---
    tags:
        - Sessions
    parameters:
      - name: session_id
        in: query
        required: false
        type: string
      - name: status
        in: query
        required: false
        type: string
    responses:
        200:
            description: The count of sessions
            schema:
                type: integer
        500:
            description: An error occurred
            schema:
                type: object
                properties:
                    message:
                        type: string
    """
    try:
        kwargs = request.args.to_dict()
        count = session_store.count(**kwargs)
        response = count
        return jsonify(response), 200
    except Exception:
        return jsonify({"message": "An unexpected error occurred"}), 500


@bp.route("/count", methods=["POST"])
@jwt_auth_required(role="admin")
def sessions_count():
    """Sessions count
    Retrieves the count of sessions based on the provided parameters.
    Works much like the GET /sessions/count endpoint, but allows for more complex queries.
    By specifying a parameter in the request body, you can filter the sessions based on that parameter,
    if the parameter value contains a comma, the filter will be an "in" query, meaning that the sessions
    will be returned if the specified field contains any of the values in the parameter.
    ---
    tags:
        - Sessions
    parameters:
      - name: session
        in: body
        required: false
        schema:
          type: object
          properties:
            status:
              type: string
            session_id:
              type: string
    responses:
        200:
            description: The count of sessions
            schema:
                type: integer
        500:
            description: An error occurred
            schema:
                type: object
                properties:
                    message:
                        type: string
    """
    try:
        kwargs = get_post_data_to_kwargs(request)
        count = session_store.count(**kwargs)
        response = count
        return jsonify(response), 200
    except Exception:
        return jsonify({"message": "An unexpected error occurred"}), 500


@bp.route("/<string:id>", methods=["GET"])
@jwt_auth_required(role="admin")
def get_session(id: str):
    """Get session
    Retrieves a session based on the provided id.
    ---
    tags:
        - Sessions
    parameters:
      - name: id
        in: path
        required: true
        type: string
        description: The id of the session
    responses:
        200:
            description: The session
            schema:
                $ref: '#/definitions/Session'
        404:
            description: The session was not found
            schema:
                type: object
                properties:
                    error:
                        type: string
        500:
            description: An error occurred
            schema:
                type: object
                properties:
                    message:
                        type: string
    """
    try:
        session = session_store.get(id, use_typing=False)
        response = session

        return jsonify(response), 200
    except EntityNotFound:
        return jsonify({"message": f"Entity with id: {id} not found"}), 404
    except Exception:
        return jsonify({"message": "An unexpected error occurred"}), 500


@bp.route("/", methods=["POST"])
@jwt_auth_required(role="admin")
def post():
    """Create session
    Creates a new session based on the provided data.
    ---
    tags:
        - Sessions
    parameters:
      - name: session
        in: body
        required: true
        schema:
          type: object
          properties:
            session_id:
              type: string
            session_config:
              type: object
    responses:
        201:
            description: The created session
            schema:
                $ref: '#/definitions/Session'
        500:
            description: An error occurred
            schema:
                type: object
                properties:
                    message:
                        type: string
    """
    try:
        data = request.json if request.headers["Content-Type"] == "application/json" else request.form.to_dict()
        successful, result = session_store.add(data)
        response = result
        status_code: int = 201 if successful else 400

        return jsonify(response), status_code
    except Exception:
        return jsonify({"message": "An unexpected error occurred"}), 500


def _get_number_of_available_clients():
    result = 0
    for combiner in control.network.get_combiners():
        try:
            nr_active_clients = len(combiner.list_active_clients())
            result = result + int(nr_active_clients)
        except CombinerUnavailableError:
            return 0

    return result


@bp.route("/start", methods=["POST"])
@jwt_auth_required(role="admin")
def start_session():
    """Start a new session.
    param: session_id: The session id to start.
    type: session_id: str
    param: rounds: The number of rounds to run.
    type: rounds: int
    """
    try:
        data = request.json if request.headers["Content-Type"] == "application/json" else request.form.to_dict()
        session_id: str = data.get("session_id")
        rounds: int = data.get("rounds", "")
        round_timeout: int = data.get("round_timeout", None)

        if not session_id or session_id == "":
            return jsonify({"message": "Session ID is required"}), 400

        session = session_store.get(session_id, use_typing=False)

        session_config = session["session_config"]
        model_id = session_config["model_id"]
        min_clients = session_config["clients_required"]

        if control.state() == ReducerState.monitoring:
            return jsonify({"message": "A session is already running."})

        if not rounds or not isinstance(rounds, int):
            rounds = session_config["rounds"]
        nr_available_clients = _get_number_of_available_clients()

        if nr_available_clients < min_clients:
            return jsonify({"message": f"Number of available clients is lower than the required minimum of {min_clients}"}), 400

        _ = model_store.get(model_id, use_typing=False)

        threading.Thread(target=control.start_session, args=(session_id, rounds, round_timeout)).start()

        return jsonify({"message": "Session started"}), 200
    except Exception:
        return jsonify({"message": "An unexpected error occurred"}), 500

@bp.route("/<string:id>", methods=["PATCH"])
@jwt_auth_required(role="admin")
def patch_session(id: str):
    """Patch session
    Updates a session based on the provided id. Only the fields that are present in the request will be updated.
    ---
    tags:
        - Sessions
    parameters:
        - name: id
            in: path
            required: true
            type: string
            description: The id or session property of the session
        - name: session
            in: body
            required: true
            type: object
            description: The session data to update
    responses:
        200:
            description: The updated session
            schema:
                $ref: '#/definitions/Session'
        404:
            description: The session was not found
            schema:
                type: object
                properties:
                    message:
                        type: string
        500:
            description: An error occurred
            schema:
                type: object
                properties:
                    message:
                        type: string
    """
    try:
        session = session_store.get(id, use_typing=False)

        data = request.get_json()
        _id = session["id"]

        # Update the session with the new data
        # Only update the fields that are present in the request
        for key, value in data.items():
            if key in ["_id", "session_id"]:
                continue
            session[key] = value

        success, message = session_store.update(_id, session)

        if success:
            response = session
            return jsonify(response), 200

        return jsonify({"message": f"Failed to update session: {message}"}), 500
    except EntityNotFound:
        return jsonify({"message": f"Entity with id: {id} not found"}), 404
    except Exception:
        return jsonify({"message": "An unexpected error occurred"}), 500


@bp.route("/<string:id>", methods=["PUT"])
@jwt_auth_required(role="admin")
def put_session(id: str):
    """Put session
    Updates a session based on the provided id. All fields will be updated with the new data.
    ---
    tags:
        - Sessions
    parameters:
        - name: id
            in: path
            required: true
            type: string
            description: The id or session property of the session
        - name: session
            in: body
            required: true
            type: object
            description: The session data to update
    responses:
        200:
            description: The updated session
            schema:
                $ref: '#/definitions/Session'
        404:
            description: The session was not found
            schema:
                type: object
                properties:
                    message:
                        type: string
        500:
            description: An error occurred
            schema:
                type: object
                properties:
                    message:
                        type: string
    """
    try:
        session = session_store.get(id, use_typing=False)
        data = request.get_json()
        _id = session["id"]

        success, message = session_store.update(_id, data)

        if success:
            response = session
            return jsonify(response), 200

        return jsonify({"message": f"Failed to update session: {message}"}), 500
    except EntityNotFound:
        return jsonify({"message": f"Entity with id: {id} not found"}), 404
    except Exception:
        return jsonify({"message": "An unexpected error occurred"}), 500



from typing import Any, Dict, List, Tuple

import pymongo
from pymongo.database import Database

from fedn.network.storage.statestore.stores.store import Store


class Status:
    def __init__(
            self,
            id: str,
            status: str,
            timestamp: str,
            log_level: str,
            data: str,
            correlation_id: str,
            type: str,
            extra: str,
            session_id: str,
            sender: dict = None
    ):
        self.id = id
        self.status = status
        self.timestamp = timestamp
        self.log_level = log_level
        self.data = data
        self.correlation_id = correlation_id
        self.type = type
        self.extra = extra
        self.session_id = session_id
        self.sender = sender

    def from_dict(data: dict) -> "Status":
        return Status(
            id=str(data["_id"]),
            status=data["status"] if "status" in data else None,
            timestamp=data["timestamp"] if "timestamp" in data else None,
            log_level=data["logLevel"] if "logLevel" in data else None,
            data=data["data"] if "data" in data else None,
            correlation_id=data["correlationId"] if "correlationId" in data else None,
            type=data["type"] if "type" in data else None,
            extra=data["extra"] if "extra" in data else None,
            session_id=data["sessionId"] if "sessionId" in data else None,
            sender=data["sender"] if "sender" in data else None
        )


class StatusStore(Store[Status]):
    def __init__(self, database: Database, collection: str):
        super().__init__(database, collection)

    def get(self, id: str, use_typing: bool = False) -> Status:
        """Get an entity by id
        param id: The id of the entity
            type: str
            description: The id of the entity, can be either the id or the status (property)
        param use_typing: Whether to return the entity as a typed object or as a dict
            type: bool
        return: The entity
        """
        response = super().get(id, use_typing=use_typing)
        return Status.from_dict(response) if use_typing else response

    def update(self, id: str, item: Status) -> bool:
        raise NotImplementedError("Update not implemented for StatusStore")

    def add(self, item: Status)-> Tuple[bool, Any]:
        raise NotImplementedError("Add not implemented for StatusStore")

    def delete(self, id: str) -> bool:
        raise NotImplementedError("Delete not implemented for StatusStore")

    def list(self, limit: int, skip: int, sort_key: str, sort_order=pymongo.DESCENDING, use_typing: bool = False, **kwargs) -> Dict[int, List[Status]]:
        """List entities
        param limit: The maximum number of entities to return
            type: int
            description: The maximum number of entities to return
        param skip: The number of entities to skip
            type: int
            description: The number of entities to skip
        param sort_key: The key to sort by
            type: str
            description: The key to sort by
        param sort_order: The order to sort by
            type: pymongo.DESCENDING
            description: The order to sort by
        param use_typing: Whether to return the entities as typed objects or as dicts
            type: bool
            description: Whether to return the entities as typed objects or as dicts.
        """
        response = super().list(limit, skip, sort_key or "timestamp", sort_order, use_typing=use_typing, **kwargs)

        result = [Status.from_dict(item) for item in response["result"]] if use_typing else response["result"]

        return {"count": response["count"], "result": result}



# This file contains the Connector class for announcing combiner to the FEDn network via the discovery service (REST-API).
# The Connector class is used by the Combiner class in fedn/network/combiner/server.py.
# Once announced, the combiner will be able to receive controller requests from the controllerStub via gRPC.
# The discovery service will also add the combiner to the statestore.
#
#
import enum
import os

import requests

from fedn.common.log_config import logger


class Status(enum.Enum):
    """Enum for representing the status of a combiner announcement."""

    Unassigned = 0
    Assigned = 1
    TryAgain = 2
    UnAuthorized = 3
    UnMatchedConfig = 4


class ConnectorCombiner:
    """Connector for annnouncing combiner to the FEDn network.

    :param host: host of discovery service
    :type host: str
    :param port: port of discovery service
    :type port: int
    :param myhost: host of combiner
    :type myhost: str
    :param fqdn: fully qualified domain name of combiner
    :type fqdn: str
    :param myport: port of combiner
    :type myport: int
    :param token: token for authentication
    :type token: str
    :param name: name of combiner
    :type name: str
    :param secure: True if https is used, False if http
    :type secure: bool
    :param verify: True if certificate is verified, False if not
    :type verify: bool
    """

    def __init__(self, host, port, myhost, fqdn, myport, token, name, secure=False, verify=False):
        """Initialize the ConnectorCombiner.

        :param host: The host of the discovery service.
        :type host: str
        :param port: The port of the discovery service.
        :type port: int
        :param myhost: The host of the combiner.
        :type myhost: str
        :param fqdn: The fully qualified domain name of the combiner.
        :type fqdn: str
        :param myport: The port of the combiner.
        :type myport: int
        :param token: The token for the discovery service.
        :type token: str
        :param name: The name of the combiner.
        :type name: str
        :param secure: Use https for the connection to the discovery service.
        :type secure: bool
        :param verify: Verify the connection to the discovery service.
        :type verify: bool
        """
        self.host = host
        self.fqdn = fqdn
        self.port = port
        self.myhost = myhost
        self.myport = myport
        self.token = token
        self.token_scheme = os.environ.get("FEDN_AUTH_SCHEME", "Bearer")
        self.name = name
        self.secure = secure
        self.verify = verify

        if not self.token:
            self.token = os.environ.get("FEDN_AUTH_TOKEN", None)

        # for https we assume a an ingress handles permanent redirect (308)
        self.prefix = "http://"
        if port:
            self.connect_string = "{}{}:{}".format(self.prefix, self.host, self.port)
        else:
            self.connect_string = "{}{}".format(self.prefix, self.host)

        logger.info("Setting connection string to {}".format(self.connect_string))

    def announce(self):
        """Announce combiner to FEDn network via discovery service (REST-API).

        :return: Tuple with announcement Status, FEDn network configuration if sucessful, else None.
        :rtype: :class:`fedn.network.combiner.connect.Status`, str
        """
        payload = {"combiner_id": self.name, "address": self.myhost, "fqdn": self.fqdn, "port": self.myport, "secure_grpc": self.secure}
        url_prefix = os.environ.get("FEDN_CUSTOM_URL_PREFIX", "")
        try:
            retval = requests.post(
                self.connect_string + url_prefix + "/add_combiner",
                json=payload,
                verify=self.verify,
                headers={"Authorization": f"{self.token_scheme} {self.token}"},
            )
        except Exception:
            return Status.Unassigned, {}

        if retval.status_code == 400:
            # Get error messange from response
            reason = retval.json()["message"]
            return Status.UnMatchedConfig, reason

        if retval.status_code == 401:
            reason = "Unauthorized connection to reducer, make sure the correct token is set"
            return Status.UnAuthorized, reason

        if retval.status_code >= 200 and retval.status_code < 204:
            if retval.json()["status"] == "retry":
                reason = retval.json()["message"]
                return Status.TryAgain, reason
            return Status.Assigned, retval.json()

        return Status.Unassigned, None



from flask import Blueprint, jsonify, request

from fedn.network.api.auth import jwt_auth_required
from fedn.network.api.v1.shared import api_version, get_post_data_to_kwargs, get_typed_list_headers, mdb
from fedn.network.storage.statestore.stores.combiner_store import CombinerStore
from fedn.network.storage.statestore.stores.shared import EntityNotFound

bp = Blueprint("combiner", __name__, url_prefix=f"/api/{api_version}/combiners")

combiner_store = CombinerStore(mdb, "network.combiners")


@bp.route("/", methods=["GET"])
@jwt_auth_required(role="admin")
def get_combiners():
    """Get combiners
    Retrieves a list of combiners based on the provided parameters.
    By specifying a parameter in the url, you can filter the combiners based on that parameter,
    and the response will contain only the combiners that match the filter.
    ---
    tags:
        - Combiners
    parameters:
      - name: name
        in: query
        required: false
        type: string
        description: The name of the combiner
      - name: address
        in: query
        required: false
        type: string
      - name: ip
        in: query
        required: false
        type: string
        description: The ip of the combiner
      - name: X-Limit
        in: header
        required: false
        type: integer
        description: The maximum number of combiners to retrieve
      - name: X-Skip
        in: header
        required: false
        type: integer
        description: The number of combiners to skip
      - name: X-Sort-Key
        in: header
        required: false
        type: string
        description: The key to sort the combiners by
      - name: X-Sort-Order
        in: header
        required: false
        type: string
        description: The order to sort the combiners in ('asc' or 'desc')
    definitions:
      Combiner:
        type: object
        properties:
          id:
            type: string
          name:
            type: string
          address:
            type: string
          config:
            type: object
          fqdn:
            type: string
            description: Fully Qualified Domain Name (FQDN)
          ip:
            type: string
          key:
            type: string
          parent:
            type: object
          port:
            type: integer
          updated_at:
            type: string
            format: date-time
    responses:
      200:
        description: A list of combiners and the total count.
        schema:
            type: object
            properties:
                count:
                    type: integer
                result:
                    type: array
                    items:
                        $ref: '#/definitions/Combiner'
      500:
        description: An error occurred
        schema:
            type: object
            properties:
                message:
                    type: string
    """
    try:
        limit, skip, sort_key, sort_order, _ = get_typed_list_headers(request.headers)

        kwargs = request.args.to_dict()

        combiners = combiner_store.list(limit, skip, sort_key, sort_order, use_typing=False, **kwargs)

        result = combiners["result"]

        response = {"count": combiners["count"], "result": result}

        return jsonify(response), 200
    except Exception:
        return jsonify({"message": "An unexpected error occurred"}), 500


@bp.route("/list", methods=["POST"])
@jwt_auth_required(role="admin")
def list_combiners():
    """List combiners
    Retrieves a list of combiners based on the provided parameters.
    Works much like the GET /combiners endpoint, but allows for more complex queries.
    By specifying a parameter in the request body, you can filter the combiners based on that parameter,
    and the response will contain only the combiners that match the filter. If the parameter value contains a comma,
    the filter will be an "in" query, meaning that the combiners will be returned if the specified field contains any of the values in the parameter.
    ---
    tags:
        - Combiners
    parameters:
      - name: combiner
        in: body
        required: false
        type: object
        description: Object containing the parameters to filter the combiners by
        schema:
          type: object
          properties:
            name:
              type: string
            address:
              type: string
            ip:
              type: string
      - name: X-Limit
        in: header
        required: false
        type: integer
        description: The maximum number of combiners to retrieve
      - name: X-Skip
        in: header
        required: false
        type: integer
        description: The number of combiners to skip
      - name: X-Sort-Key
        in: header
        required: false
        type: string
        description: The key to sort the combiners by
      - name: X-Sort-Order
        in: header
        required: false
        type: string
        description: The order to sort the combiners in ('asc' or 'desc')
    responses:
      200:
        description: A list of combiners and the total count.
        schema:
            type: object
            properties:
                count:
                    type: integer
                result:
                    type: array
                    items:
                        $ref: '#/definitions/Combiner'
      500:
        description: An error occurred
        schema:
            type: object
            properties:
                message:
                    type: string
    """
    try:
        limit, skip, sort_key, sort_order, _ = get_typed_list_headers(request.headers)

        kwargs = get_post_data_to_kwargs(request)

        combiners = combiner_store.list(limit, skip, sort_key, sort_order, use_typing=False, **kwargs)

        result = combiners["result"]

        response = {"count": combiners["count"], "result": result}

        return jsonify(response), 200
    except Exception:
        return jsonify({"message": "An unexpected error occurred"}), 500


@bp.route("/count", methods=["GET"])
@jwt_auth_required(role="admin")
def get_combiners_count():
    """Combiners count
    Retrieves the count of combiners based on the provided parameters.
    By specifying a parameter in the url, you can filter the combiners based on that parameter,
    and the response will contain only the count of combiners that match the filter.
    ---
    tags:
        - Combiners
    parameters:
      - name: name
        in: query
        required: false
        type: string
        description: The name of the combiner
      - name: address
        in: query
        required: false
        type: string
      - name: ip
        in: query
        required: false
        type: string
        description: The ip of the combiner
    responses:
        200:
            description: The count of combiners
            schema:
                type: integer
        500:
            description: An error occurred
            schema:
                type: object
                properties:
                    message:
                        type: string
    """
    try:
        kwargs = request.args.to_dict()
        count = combiner_store.count(**kwargs)
        response = count
        return jsonify(response), 200
    except Exception:
        return jsonify({"message": "An unexpected error occurred"}), 500


@bp.route("/count", methods=["POST"])
@jwt_auth_required(role="admin")
def combiners_count():
    """Combiners count
    Retrieves the count of combiners based on the provided parameters.
    Works much like the GET /combiners/count endpoint, but allows for more complex queries.
    By specifying a parameter in the request body, you can filter the combiners based on that parameter,
    and the response will contain only the count of combiners that match the filter. If the parameter value contains a comma,
    the filter will be an "in" query, meaning that the combiners will be returned if the specified field contains any of the values in the parameter.
    ---
    tags:
        - Combiners
    parameters:
      - name: combiner
        in: body
        required: false
        type: object
        description: Object containing the parameters to filter the combiners by
        schema:
          type: object
          properties:
            name:
                type: string
            address:
                type: string
            ip:
                type: string
    responses:
        200:
            description: The count of combiners
            schema:
                type: integer
        500:
            description: An error occurred
            schema:
                type: object
                properties:
                    message:
                        type: string
    """
    try:
        kwargs = get_post_data_to_kwargs(request)
        count = combiner_store.count(**kwargs)
        response = count
        return jsonify(response), 200
    except Exception:
        return jsonify({"message": "An unexpected error occurred"}), 500


@bp.route("/<string:id>", methods=["GET"])
@jwt_auth_required(role="admin")
def get_combiner(id: str):
    """Get combiner
    Retrieves a combiner based on the provided id.
    ---
    tags:
        - Combiners
    parameters:
      - name: id
        in: path
        required: true
        type: string
        description: The id of the combiner
    responses:
        200:
            description: The combiner
            schema:
                $ref: '#/definitions/Combiner'
        404:
            description: The combiner was not found
            schema:
                type: object
                properties:
                    error:
                        type: string
        500:
            description: An error occurred
            schema:
                type: object
                properties:
                    message:
                        type: string
    """
    try:
        combiner = combiner_store.get(id, use_typing=False)
        response = combiner

        return jsonify(response), 200
    except EntityNotFound:
        return jsonify({"message": f"Entity with id: {id} not found"}), 404
    except Exception:
        return jsonify({"message": "An unexpected error occurred"}), 500



import click

from .main import main


@main.group("controller")
@click.pass_context
def controller_cmd(ctx):
    """:param ctx:"""
    pass


@controller_cmd.command("start")
@click.pass_context
def controller_cmd(ctx):
    from fedn.network.api.server import start_server_api

    start_server_api()



""" The FEDn client package is responsible for executing the federated learning tasks, including ML model training and validation. It's the acting gRPC client for the federated network.
The client first connacts the centralized controller to receive :class:`fedn.network.combiner.Combiner` assingment. The client then connects to the combiner and
sends requests to the combiner to receive model updates and send model updates."""
# flake8: noqa



import os

import yaml

from fedn.utils.dist import get_package_path

SECRET_KEY = os.environ.get("FEDN_JWT_SECRET_KEY", False)
FEDN_JWT_CUSTOM_CLAIM_KEY = os.environ.get("FEDN_JWT_CUSTOM_CLAIM_KEY", False)
FEDN_JWT_CUSTOM_CLAIM_VALUE = os.environ.get("FEDN_JWT_CUSTOM_CLAIM_VALUE", False)

FEDN_AUTH_WHITELIST_URL_PREFIX = os.environ.get("FEDN_AUTH_WHITELIST_URL_PREFIX", False)
FEDN_JWT_ALGORITHM = os.environ.get("FEDN_JWT_ALGORITHM", "HS256")
FEDN_AUTH_SCHEME = os.environ.get("FEDN_AUTH_SCHEME", "Bearer")
FEDN_AUTH_REFRESH_TOKEN_URI = os.environ.get("FEDN_AUTH_REFRESH_TOKEN_URI", False)
FEDN_AUTH_REFRESH_TOKEN = os.environ.get("FEDN_AUTH_REFRESH_TOKEN", False)
FEDN_CUSTOM_URL_PREFIX = os.environ.get("FEDN_CUSTOM_URL_PREFIX", "")

FEDN_PACKAGE_EXTRACT_DIR = os.environ.get("FEDN_PACKAGE_EXTRACT_DIR", "package")


def get_environment_config():
    """Get the configuration from environment variables."""
    global STATESTORE_CONFIG
    global MODELSTORAGE_CONFIG
    if not os.environ.get("STATESTORE_CONFIG", False):
        STATESTORE_CONFIG = get_package_path() + "/common/settings-controller.yaml.template"
    else:
        STATESTORE_CONFIG = os.environ.get("STATESTORE_CONFIG")

    if not os.environ.get("MODELSTORAGE_CONFIG", False):
        MODELSTORAGE_CONFIG = get_package_path() + "/common/settings-controller.yaml.template"
    else:
        MODELSTORAGE_CONFIG = os.environ.get("MODELSTORAGE_CONFIG")


def get_statestore_config(file=None):
    """Get the statestore configuration from file.

    :param file: The statestore configuration file (yaml) path (optional).
    :type file: str
    :return: The statestore configuration as a dict.
    :rtype: dict
    """
    if file is None:
        get_environment_config()
        file = STATESTORE_CONFIG
    with open(file, "r") as config_file:
        try:
            settings = dict(yaml.safe_load(config_file))
        except yaml.YAMLError as e:
            raise (e)
    return settings["statestore"]


def get_modelstorage_config(file=None):
    """Get the model storage configuration from file.

    :param file: The model storage configuration file (yaml) path (optional).
    :type file: str
    :return: The model storage configuration as a dict.
    :rtype: dict
    """
    if file is None:
        get_environment_config()
        file = MODELSTORAGE_CONFIG
    with open(file, "r") as config_file:
        try:
            settings = dict(yaml.safe_load(config_file))
        except yaml.YAMLError as e:
            raise (e)
    return settings["storage"]


def get_network_config(file=None):
    """Get the network configuration from file.

    :param file: The network configuration file (yaml) path (optional).
    :type file: str
    :return: The network id.
    :rtype: str
    """
    if file is None:
        get_environment_config()
        file = STATESTORE_CONFIG
    with open(file, "r") as config_file:
        try:
            settings = dict(yaml.safe_load(config_file))
        except yaml.YAMLError as e:
            raise (e)
    return settings["network_id"]


def get_controller_config(file=None):
    """Get the controller configuration from file.

    :param file: The controller configuration file (yaml) path (optional).
    :type file: str
    :return: The controller configuration as a dict.
    :rtype: dict
    """
    if file is None:
        get_environment_config()
        file = STATESTORE_CONFIG
    with open(file, "r") as config_file:
        try:
            settings = dict(yaml.safe_load(config_file))
        except yaml.YAMLError as e:
            raise (e)
    return settings["controller"]



from datetime import datetime
from typing import Any, Dict, List, Tuple

import pymongo
from pymongo.database import Database

from fedn.network.storage.statestore.stores.store import Store


class Client:
    def __init__(self, id: str, name: str, combiner: str, combiner_preferred: str, ip: str, status: str, updated_at: str, last_seen: datetime):
        self.id = id
        self.name = name
        self.combiner = combiner
        self.combiner_preferred = combiner_preferred
        self.ip = ip
        self.status = status
        self.updated_at = updated_at
        self.last_seen = last_seen

    def from_dict(data: dict) -> "Client":
        return Client(
            id=str(data["_id"]),
            name=data["name"] if "name" in data else None,
            combiner=data["combiner"] if "combiner" in data else None,
            combiner_preferred=data["combiner_preferred"] if "combiner_preferred" in data else None,
            ip=data["ip"] if "ip" in data else None,
            status=data["status"] if "status" in data else None,
            updated_at=data["updated_at"] if "updated_at" in data else None,
            last_seen=data["last_seen"] if "last_seen" in data else None
        )


class ClientStore(Store[Client]):
    def __init__(self, database: Database, collection: str):
        super().__init__(database, collection)

    def get(self, id: str, use_typing: bool = False) -> Client:
        """Get an entity by id
        param id: The id of the entity
            type: str
        param use_typing: Whether to return the entity as a typed object or as a dict
            type: bool
        return: The entity
        """
        response = super().get(id, use_typing=use_typing)
        return Client.from_dict(response) if use_typing else response

    def update(self, id: str, item: Client) -> bool:
        raise NotImplementedError("Update not implemented for ClientStore")

    def add(self, item: Client)-> Tuple[bool, Any]:
        raise NotImplementedError("Add not implemented for ClientStore")

    def delete(self, id: str) -> bool:
        raise NotImplementedError("Delete not implemented for ClientStore")

    def list(self, limit: int, skip: int, sort_key: str, sort_order=pymongo.DESCENDING, use_typing: bool = False, **kwargs) -> Dict[int, List[Client]]:
        """List entities
        param limit: The maximum number of entities to return
            type: int
        param skip: The number of entities to skip
            type: int
        param sort_key: The key to sort by
            type: str
        param sort_order: The order to sort by
            type: pymongo.DESCENDING | pymongo.ASCENDING
        param use_typing: Whether to return the entities as typed objects or as dicts
            type: bool
        param kwargs: Additional query parameters
            type: dict
            example: {"key": "models"}
        return: A dictionary with the count and the result
        """
        response = super().list(limit, skip, sort_key or "last_seen", sort_order, use_typing=use_typing, **kwargs)

        result = [Client.from_dict(item) for item in response["result"]] if use_typing else response["result"]

        return {
            "count": response["count"],
            "result": result
        }

    def count(self, **kwargs) -> int:
        return super().count(**kwargs)



"""Portions of this code are derived from the Apache 2.0 licensed project mlflow (https://mlflow.org/).,
with modifications made by Scaleout Systems AB.
Copyright (c) 2018 Databricks, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import os
import shutil
import sys
import tempfile
import uuid
from contextlib import contextmanager
from pathlib import Path

import yaml

from fedn.common.log_config import logger
from fedn.utils import PYTHON_VERSION
from fedn.utils.environment import _PythonEnv
from fedn.utils.process import _exec_cmd, _join_commands

_IS_UNIX = os.name != "nt"


@contextmanager
def remove_on_error(path: os.PathLike, onerror=None):
    """A context manager that removes a file or directory if an exception is raised during
    execution.
    """
    try:
        yield
    except Exception as e:
        if onerror:
            onerror(e)
        if os.path.exists(path):
            if os.path.isfile(path):
                os.remove(path)
            elif os.path.isdir(path):
                shutil.rmtree(path)
        raise


def _install_python(version, pyenv_root=None, capture_output=False):
    """Installs a specified version of python with pyenv and returns a path to the installed python
    binary.
    """
    raise NotImplementedError("This function is not implemented yet.")


def _is_virtualenv_available():
    """Returns True if virtualenv is available, otherwise False.
    """
    return shutil.which("virtualenv") is not None


def _validate_virtualenv_is_available():
    """Validates virtualenv is available. If not, throws an `Exception` with a brief instruction
    on how to install virtualenv.
    """
    if not _is_virtualenv_available():
        raise Exception("Could not find the virtualenv binary. Run `pip install virtualenv` to install " "virtualenv.")


def _get_virtualenv_extra_env_vars(env_root_dir=None):
    extra_env = {
        # PIP_NO_INPUT=1 makes pip run in non-interactive mode,
        # otherwise pip might prompt "yes or no" and ask stdin input
        "PIP_NO_INPUT": "1",
    }
    return extra_env


def _get_python_env(python_env_file):
    """Parses a python environment file and returns a dictionary with the parsed content.
    """
    if os.path.exists(python_env_file):
        return _PythonEnv.from_yaml(python_env_file)


def _create_virtualenv(python_bin_path, env_dir, python_env, extra_env=None, capture_output=False):
    # Created a command to activate the environment
    paths = ("bin", "activate") if _IS_UNIX else ("Scripts", "activate.bat")
    activate_cmd = env_dir.joinpath(*paths)
    activate_cmd = f"source {activate_cmd}" if _IS_UNIX else str(activate_cmd)

    if env_dir.exists():
        logger.info("Environment %s already exists", env_dir)
        return activate_cmd

    with remove_on_error(
        env_dir,
        onerror=lambda e: logger.warning(
            "Encountered an unexpected error: %s while creating a virtualenv environment in %s, " "removing the environment directory...",
            repr(e),
            env_dir,
        ),
    ):
        logger.info("Creating a new environment in %s with %s", env_dir, python_bin_path)
        _exec_cmd(
            [sys.executable, "-m", "virtualenv", "--python", python_bin_path, env_dir],
            capture_output=capture_output,
        )

        logger.info("Installing dependencies")
        for deps in filter(None, [python_env.build_dependencies, python_env.dependencies]):
            with tempfile.TemporaryDirectory() as tmpdir:
                tmp_req_file = f"requirements.{uuid.uuid4().hex}.txt"
                Path(tmpdir).joinpath(tmp_req_file).write_text("\n".join(deps))
                cmd = _join_commands(activate_cmd, f"python -m pip install -r {tmp_req_file}")
                _exec_cmd(cmd, capture_output=capture_output, cwd=tmpdir, extra_env=extra_env)

    return activate_cmd


def _read_yaml_file(file_path):
    try:
        cfg = None
        with open(file_path, "rb") as config_file:
            cfg = yaml.safe_load(config_file.read())

    except Exception as e:
        logger.error(f"Error trying to read yaml file: {file_path}")
        raise e
    return cfg


class Dispatcher:
    """Dispatcher class for compute packages.

    :param config: The configuration.
    :type config: dict
    :param dir: The directory to dispatch to.
    :type dir: str
    """

    def __init__(self, config, project_dir):
        """Initialize the dispatcher."""
        self.config = config
        self.project_dir = project_dir
        self.activate_cmd = ""
        self.python_env_path = ""

    def _get_or_create_python_env(self, capture_output=False, pip_requirements_override=None):
        python_env = self.config.get("python_env", "")
        if not python_env:
            logger.info("No python_env specified in the configuration, using the system Python.")
            return python_env
        else:
            python_env_path = os.path.join(self.project_dir, python_env)
            if not os.path.exists(python_env_path):
                raise Exception("Compute package specified python_env file %s, but no such " "file was found." % python_env_path)
            python_env = _get_python_env(python_env_path)

        extra_env = _get_virtualenv_extra_env_vars()
        env_dir = Path(self.project_dir) / Path(python_env.name)
        self.python_env_path = env_dir
        try:
            python_bin_path = _install_python(python_env.python, capture_output=True)
        except NotImplementedError:
            logger.warning("Failed to install Python: %s", python_env.python)
            logger.warning("Python version installation is not implemented yet.")
            logger.info(f"Using the system Python version: {PYTHON_VERSION}")
            python_bin_path = Path(sys.executable)

        try:
            activate_cmd = _create_virtualenv(
                python_bin_path,
                env_dir,
                python_env,
                extra_env=extra_env,
                capture_output=capture_output,
            )
            # Install additional dependencies specified by `requirements_override`
            if pip_requirements_override:
                logger.info("Installing additional dependencies specified by " f"pip_requirements_override: {pip_requirements_override}")
                cmd = _join_commands(
                    activate_cmd,
                    f"python -m pip install --quiet -U {' '.join(pip_requirements_override)}",
                )
                _exec_cmd(cmd, capture_output=capture_output, extra_env=extra_env)
            self.activate_cmd = activate_cmd
            return activate_cmd
        except Exception:
            logger.critical("Encountered unexpected error while creating %s", env_dir)
            if env_dir.exists():
                logger.warning("Attempting to remove %s", env_dir)
                shutil.rmtree(env_dir, ignore_errors=True)
                msg = "Failed to remove %s" if env_dir.exists() else "Successfully removed %s"
                logger.warning(msg, env_dir)

            raise

    def run_cmd(self, cmd_type, capture_output=False, extra_env=None, synchronous=True, stream_output=False):
        """Run a command.

        :param cmd_type: The command type.
        :type cmd_type: str
        :return:
        """
        try:
            cmdsandargs = cmd_type.split(" ")

            entry_point = self.config["entry_points"][cmdsandargs[0]]["command"]

            # remove the first element,  that is not a file but a command
            args = cmdsandargs[1:]

            # Join entry point and arguments into a single command as a string
            entry_point_args = " ".join(args)
            entry_point = f"{entry_point} {entry_point_args}"

            if self.activate_cmd:
                cmd = _join_commands(self.activate_cmd, entry_point)
            else:
                cmd = _join_commands(entry_point)

            logger.info("Running command: {}".format(cmd))
            _exec_cmd(
                cmd,
                cwd=self.project_dir,
                throw_on_error=True,
                extra_env=extra_env,
                capture_output=capture_output,
                synchronous=synchronous,
                stream_output=stream_output,
            )

            logger.info("Done executing {}".format(cmd_type))
        except IndexError:
            message = "No such argument or configuration to run."
            logger.error(message)



""" The statestore package is responsible for storing various states of the federated network. Such as announced combiners and assigned clients. It also stores metadata about
models, rounds, sessions, compute packages and model validations. """
# flake8: noqa
